/*
 * AppliedMicro X-Gene PCIe DMA Driver
 *
 * Copyright (c) 2014, Applied Micro Circuits Corporation
 * Author: Tanmay Inamdar <tinamdar@apm.com>
 *         Mayuresh Chitale <mchitale@apm.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/pm_runtime.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/platform_device.h>
#include <linux/of.h>
#include <linux/of_irq.h>
#include <linux/of_device.h>
#include <linux/delay.h>
#include <asm/cacheflush.h>
#include "xgene-pcie-dma.h"

static struct list_head dma_port_list;
/* Enable polling instead of irq. Seeing issue with interrupts */
int enable_polling = 0;
/* This will flush only the descriptors and not data buffers */
int enable_flush = 0;
extern u32 xgene_pcie_dma_read_intr(int port);
extern void xgene_pcie_dma_write_intr(int port, u32 val);
extern void __iomem * xgene_pcie_dma_cfg_base (int port);
static struct xgene_dma_port *port_g = NULL;
struct xgene_pcie_info {
	u8	max_chan;
	u32	src_elems;
	u32	dst_elems;
	u32	sts_elems;
};

#define INFO(_max_chan, _src_elems, _dst_elems, _sts_elems)	\
	(kernel_ulong_t)&(struct xgene_pcie_info) {  		\
	 .max_chan = (_max_chan),                        	\
	 .src_elems = (_src_elems),                          	\
	 .dst_elems = (_dst_elems),                   		\
	 .sts_elems = (_sts_elems),                      	\
	 }

/* Hack since IO coherency is not working */
static void flush_to_ddr(void *addr, int len)
{
#ifdef CONFIG_DMA_FLUSH
	if (enable_flush)
		__flush_dcache_area(addr, len);
#endif
}

static inline void xgene_dma_read(const volatile void __iomem * offset, u32 *val)
{
	*val = readl(offset);
	pr_debug("DMA_READ:  Offset: 0x%p value: 0x%08X\n", offset,
			*val);
}

static inline void xgene_dma_write(volatile void __iomem * offset, u32 val)
{
	pr_debug("DMA_WRITE: Offset: 0x%p value: 0x%08X\n", offset, val);
	writel(val, offset);
}

static inline int is_dma_channel_present(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read((const volatile void __iomem *)(chan->base + DMA_CONTROL), &val);
	return val & DMA_PRESENT_MASK;
}

static inline int is_dma_channel_enabled(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	return val & DMA_ENABLE_MASK;
}

static inline int is_dma_channel_running(struct xgene_dma_chan *chan)
{
	u32 val;
	int timeout = 1000;

	/* Give DMA enought time to complete outstanding requests */
	do {
		usleep_range(10, 10);
		xgene_dma_read(chan->base + DMA_CONTROL, &val);
	} while((val & DMA_RUNNING_MASK) && timeout--);

	return (timeout == -1) ? 1: 0;
}

static inline void reset_dma_channel(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	val |= DMA_RESET_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
	/* wait 256nS */
	udelay(1);
	val &= ~DMA_RESET_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
}

static inline void disable_dma_channel(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	val &= ~DMA_ENABLE_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
}

static inline void enable_dma_channel(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	val |= DMA_ENABLE_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
}

static inline void disable_axi_dma_interrupts(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	val &= ~AXI_INTERRUPT_ENABLE_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
}

static inline void enable_axi_dma_interrupts(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	val |= AXI_INTERRUPT_ENABLE_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
}

static inline void disable_pci_dma_interrupts(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	val &= ~PCIE_INTERRUPT_ENABLE_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
}

static inline void enable_pci_dma_interrupts(struct xgene_dma_chan *chan)
{
	u32 val;
	xgene_dma_read(chan->base + DMA_CONTROL, &val);
	val |= PCIE_INTERRUPT_ENABLE_MASK;
	xgene_dma_write(chan->base + DMA_CONTROL, val);
}

static inline void xgene_dma_setup_src_q_desc(struct src_queue_desc *src_desc,
				     	    u64 addr, u32 byte_cnt, bool eop,
					    bool intr, u8 location)
{
	pr_debug("src_desc : 0x%p, addr : 0x%08llx, byte_cnt : 0x%x eop:%d, intr:%d\n",
		   src_desc, (long long unsigned int)addr, byte_cnt, eop, intr);
	src_desc->addr = addr;
	src_desc->fields.flags.byte_cnt = byte_cnt;
	src_desc->fields.flags.location = location;
	src_desc->fields.flags.eop = eop;
	src_desc->fields.flags.intr = intr;
	/* Enable cache coherency setting */
	src_desc->fields.flags.attr = 2;
	flush_to_ddr(src_desc, sizeof(struct src_queue_desc));
#ifdef XGENE_DMA_DEBUG
	print_hex_dump(KERN_DEBUG, "SRC_DESC: ", DUMP_PREFIX_NONE, 16, 1,
			src_desc, sizeof(struct src_queue_desc), 1);
#endif
}

static inline void xgene_dma_setup_dst_q_desc(struct dst_queue_desc *dst_desc,
					    u64 addr, u32 byte_cnt, u8 location)
{
	pr_debug("dst_desc : 0x%p, addr : 0x%08llx, byte_cnt : 0x%x\n",
		   dst_desc, (long long unsigned int)addr, byte_cnt);
	dst_desc->addr = addr;
	dst_desc->fields.flags.byte_cnt = byte_cnt;
	dst_desc->fields.flags.location = location;
	/* Enable cache coherency setting */
	dst_desc->fields.flags.attr = 2;
	flush_to_ddr(dst_desc, sizeof(struct dst_queue_desc));
#ifdef XGENE_DMA_DEBUG
	print_hex_dump(KERN_DEBUG, "DST_DESC: ", DUMP_PREFIX_NONE, 16, 1,
			dst_desc, sizeof(struct dst_queue_desc), 1);
#endif
}

static inline void xgene_dma_update_src_cnts(struct xgene_dma_chan *chan)
{
	chan->next_src_elem++;
	if (chan->next_src_elem >= chan->src_elems)
		chan->next_src_elem = 0;
}

static inline void xgene_dma_update_dst_cnts(struct xgene_dma_chan *chan)
{
	chan->next_dst_elem++;
	if (chan->next_dst_elem >= chan->dst_elems)
		chan->next_dst_elem = 0;
}

static inline void xgene_dma_update_sts_cnts(struct xgene_dma_chan *chan)
{
	chan->next_sts_elem++;
	if (chan->next_sts_elem >= chan->sts_elems)
		chan->next_sts_elem = 0;
}

static inline void xgene_dma_get_next_queues(struct xgene_dma_chan *chan,
					struct xgene_pcie_dma_desc *desc)
{
	desc->src_desc = (struct src_queue_desc *)(
			(char *)chan->src_elem_addr +
			(chan->next_src_elem * sizeof(*desc->src_desc)));

	desc->dst_desc = (struct dst_queue_desc *)(
			(char *)chan->dst_elem_addr +
			(chan->next_dst_elem * sizeof(*desc->dst_desc)));

	desc->sts_desc = (struct sts_queue_desc *)(
			(char *)chan->sts_elem_addr +
			(chan->next_sts_elem * sizeof(*desc->sts_desc)));
}

/**
 * DMA Helper Functions
 */

/**
 * xgene_desc_get	- get a descriptor
 * @chan: dma channel for which descriptor is required
 *
 * Obtain a descriptor for the channel. Returns NULL if none are free.
 * Once the descriptor is returned it is private until put on another
 * list or freed
 */
static struct xgene_pcie_dma_desc *xgene_desc_get(struct xgene_dma_chan *chan)
{
	struct xgene_pcie_dma_desc *desc, *_desc;
	struct xgene_pcie_dma_desc *ret = NULL;

	spin_lock_bh(&chan->lock);
	if (list_empty(&chan->descs))
		goto out;

	list_for_each_entry_safe(desc, _desc, &chan->descs, desc_node) {
		//dev_dbg(chan->dev, "Traverse free list desc = %p\n", desc);
		if (!atomic_read(&desc->busy)) {
			atomic_set(&desc->busy, 1);
			ret = desc;
			dev_dbg(chan->dev, "Found desc\n");
			break;
		}
	}
out:
	spin_unlock_bh(&chan->lock);
	return ret;
}

/**
 * xgene_desc_put	- put a descriptor
 * @chan: dma channel for which descriptor is required
 * @desc: descriptor to put
 *
 * Return a descriptor back to the free pool
 * Call with chan lock held
 */
static void xgene_desc_put(struct xgene_dma_chan *chan,
			  struct xgene_pcie_dma_desc *desc)
{
#if 0
	if (desc) {
		dev_dbg(chan->dev, "add to free list - desc = %p\n", desc);
		spin_lock_bh(&chan->lock);
		list_add_tail(&desc->desc_node, &chan->descs);
		spin_unlock_bh(&chan->lock);
	}
#endif
	atomic_set(&desc->busy, 0);
}


/**
 * xgene_start_dma - begin a DMA transaction
 * @chan: channel for which txn is to be started
 * @first: first descriptor of series
 *
 * Load a transaction into the engine. This must be called with chan->lock
 * held and bh disabled.
 */
static void xgene_start_dma(struct xgene_dma_chan *chan)
{
	u32 status, imask;
	/*write registers and en*/
	spin_lock_bh(&chan->lock);
	imask = xgene_pcie_dma_read_intr(chan->port_num);
	imask &= ~(1 << 8);
	xgene_pcie_dma_write_intr(chan->port_num, imask);
	xgene_dma_read(chan->base + DMA_CONTROL, &status);
	xgene_dma_write(chan->base + SRC_Q_LIMIT,
		      chan->next_src_elem);
	xgene_dma_write(chan->base + DST_Q_LIMIT,
		      chan->next_dst_elem);
	xgene_dma_write(chan->base + STA_Q_LIMIT,
		      chan->next_sts_elem);

	chan->state = CHANNEL_BUSY;
	dev_dbg (chan->dev, "src = %d, dst = %d, sts = %d\n",
		      chan->next_src_elem,
		      chan->next_dst_elem,
		      chan->next_sts_elem);
	spin_unlock_bh(&chan->lock);
	/*XXX: DMA channel should always be enabled */
}

/**
 * xgene_descriptor_complete - process completed descriptor
 * @chan: channel owning the descriptor
 * @desc: the descriptor itself
 *
 * Process a completed descriptor and perform any callbacks upon
 * the completion. The completion handling drops the lock during the
 * callbacks but must be called with the lock held.
 */
static void xgene_descriptor_complete(struct xgene_dma_chan *chan,
				     struct xgene_pcie_dma_desc *desc)
{
	struct dma_async_tx_descriptor	*txd = &desc->txd;
	struct sts_queue_desc *sts_desc = desc->sts_desc;
	dma_async_tx_callback callback_txd = NULL;
	void *param_txd = NULL;

	BUG_ON(!chan);
	BUG_ON(!desc);
	BUG_ON(!sts_desc);
	BUG_ON(!txd);


	(void)readl(chan->dma_sync);
	flush_to_ddr(desc->sts_desc, sizeof(struct sts_queue_desc));
#ifdef XGENE_DMA_DEBUG
        print_hex_dump(KERN_DEBUG, "STS_DESC: ", DUMP_PREFIX_NONE, 16, 1,
                        sts_desc, sizeof(struct sts_queue_desc), 1);
#endif
	if (!sts_desc->fields.dword) {
		dev_err(chan->dev, "CHAN%d: DMA transaction pending\n",  chan->id);
		goto skip_cleanup;
	}
	chan->completed = txd->cookie;
	callback_txd = txd->callback;
	param_txd = txd->callback_param;

	if (desc->flags & DMA_COMPL_SRC_UNMAP_SINGLE) {
		dev_dbg(chan->dev, "unmap source addr = 0x%llx\n", desc->src_desc->addr);
		dma_unmap_single(chan->dev, desc->src_desc->addr, desc->len, DMA_TO_DEVICE);
	}
	if (desc->flags & DMA_COMPL_DEST_UNMAP_SINGLE) {
		dev_dbg(chan->dev, "unmap dest addr = 0x%llx\n", desc->dst_desc->addr);
		dma_unmap_single(chan->dev, desc->dst_desc->addr, desc->len, DMA_FROM_DEVICE);
	}

	if (callback_txd) {
		dev_dbg(chan->dev, "TXD callback set ... calling\n");
		callback_txd(param_txd);
	}

	if (sts_desc->fields.flags.completed) {
		dev_dbg(chan->dev, "CHAN%d: DMA transaction comepleted\n",
				chan->id);
		desc->status = DMA_SUCCESS;
		chan->transfer_cnt++;
	} else {
		dev_err(chan->dev, "CHAN%d: DMA transaction error 0x%08X\n",
				chan->id, sts_desc->fields.dword);
		desc->status = DMA_ERROR;
		chan->err_cnt++;
	}
	/* Clear the status descriptor */
	sts_desc->fields.dword = 0;
	/* Freeup the elements */
	xgene_desc_put(chan, desc);

skip_cleanup:
	chan->state = CHANNEL_READY;
}

/**
 * xgene_scan_descriptors - check the descriptors in channel
 *			  mark completed when tx is completete
 * @chan: channel to scan
 *
 * Walk the descriptor chain for the device and process any entries
 * that are complete.
 */
static void xgene_scan_descriptors(struct xgene_dma_chan *chan)
{
	struct xgene_pcie_dma_desc *desc = NULL, *_desc = NULL;
	u32 imask;

	/*tx is complete*/
	list_for_each_entry_safe(desc, _desc, &chan->descs,
				 desc_node) {
		if (desc->status == DMA_IN_PROGRESS) {
			dev_dbg(chan->dev, "found %p\n",desc);
			if (!desc->sts_desc->fields.dword) {
				dev_dbg(chan->dev, "CHAN%d: DMA transaction pending\n",  chan->id);
				imask = xgene_pcie_dma_read_intr(chan->port_num);
				imask &= ~(1 << 8);
				xgene_pcie_dma_write_intr(chan->port_num, imask);
				continue;
			} else
				xgene_descriptor_complete(chan, desc);
		}
	}
	return;
}

/**
 * xgene_pcie_dma_tx_submit -	callback to submit DMA transaction
 * @tx: dma engine descriptor
 *
 * Submit the DMA trasaction for this descriptor, start if ch idle
 */
static dma_cookie_t xgene_pcie_dma_tx_submit(struct dma_async_tx_descriptor *tx)
{
	struct xgene_pcie_dma_desc *desc = to_xgene_pcie_dma_desc(tx);
	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(tx->chan);
	dma_cookie_t cookie;

	dev_dbg(chan->dev, "addr = %p\n",(void *)desc->src_desc->addr);
	spin_lock_bh(&chan->lock);
	cookie = chan->chan.cookie;
	if (++cookie < 0)
		cookie = 1;
	chan->chan.cookie = cookie;
	desc->txd.cookie = cookie;
	dev_dbg(chan->dev, "desc = %p add to active list\n", desc);
	desc->status = DMA_IN_PROGRESS;
	spin_unlock_bh(&chan->lock);

	return cookie;
}

/**
 * xgene_pcie_dma_alloc_chan_resources - Allocate dma resources
 * @chan: chan requiring attention
 *
 * Allocates DMA resources on this chan
 * Return the descriptors allocated
 */
static int xgene_pcie_dma_alloc_chan_resources(struct dma_chan *dchan)
{
	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);
	struct xgene_pcie_dma_desc *desc;
	dma_addr_t phys;
	int i = 0;

	spin_lock_bh(&chan->lock);
	if (chan->state == CHANNEL_DISABLED) {
		dev_err(chan->dev, "channel %d is DISABLED\n",chan->id);
		return -ENXIO;
	}

	chan->completed = dchan->cookie = 1;

	for (i = 0; i < DESCS_PER_CHANNEL; i++) {
		desc = (struct xgene_pcie_dma_desc *)
			dmam_alloc_coherent(chan->dev, sizeof(*desc),
					     &phys, GFP_ATOMIC);
		if(!desc) {
			dev_err (chan->dev,
					"Failure allocating channel descriptors\n");
			return -ENOMEM;
		}
		dev_dbg(chan->dev, "desc : 0x%p , phys : 0x%08lx\n",
			    desc, (unsigned long) phys);
		dma_async_tx_descriptor_init(&desc->txd, dchan);
		desc->txd.tx_submit = xgene_pcie_dma_tx_submit;
		desc->txd.phys = phys;
		atomic_set(&desc->busy, 0);
		list_add_tail(&desc->desc_node, &chan->descs);
	}
	enable_axi_dma_interrupts(chan);
	chan->descs_allocated = i;
	dev_dbg(chan->dev, "Allocated %d descriptors for XGENE DMA Chan%d\n", i,
		   chan->id);
	spin_unlock_bh(&chan->lock);
	return i;
}

/**
 * xgene_pcie_dma_free_chan_resources -  Frees dma resources
 * @chan: chan requiring attention
 *
 * Frees the allocated resources on this DMA chan
 */
static void xgene_pcie_dma_free_chan_resources(struct dma_chan *dchan)
{
	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);
	struct xgene_pcie_dma_desc *desc, *_desc;

	if (chan->state == CHANNEL_BUSY) { 
		dev_err(chan->dev, "DMA channel %d BUST\n",chan->id);
		return;
	}

	spin_lock_bh(&chan->lock);
	chan->descs_allocated = 0;
	chan->state = CHANNEL_DISABLED;	
	disable_axi_dma_interrupts(chan);

	list_for_each_entry_safe(desc, _desc, &chan->descs,
				 desc_node) {
		dev_dbg(chan->dev, "Freeing Free list Desc\n"
				"desc : 0x%p , phys : 0x%08lx\n",
			    desc, (unsigned long) desc->txd.phys);
		dmam_free_coherent(chan->dev,  sizeof(*desc), desc, desc->txd.phys);
	}

	list_for_each_entry_safe(desc, _desc, &chan->queue, desc_node) {
		dmam_free_coherent(chan->dev, sizeof(*desc), desc,
				    desc->txd.phys);
	}
	chan->state = CHANNEL_READY;
	spin_unlock_bh(&chan->lock);
}

/**
 * xgene_pcie_dma_device_control - DMA device control
 * @chan: chan for DMA control
 * @cmd: control cmd
 * @arg: cmd arg value
 *
 * Perform DMA control command
 */
static int xgene_pcie_dma_device_control(struct dma_chan *dchan,
		enum dma_ctrl_cmd cmd, unsigned long arg)
{
	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);
	struct dma_slave_config *slave = (struct dma_slave_config *)arg;

	BUG_ON(!chan);

	/* Right now we only support DMA_SLAVE_CONFIG */
	if(cmd != DMA_SLAVE_CONFIG)
		return -ENXIO;
	BUG_ON(!slave);

	chan->dma_slave = slave;
	return 0;
}

static struct dma_async_tx_descriptor *xgene_pcie_dma_prep_sg(
		struct dma_chan *dchan,
		struct scatterlist *dst_sg, unsigned int dst_nents,
		struct scatterlist *src_sg, unsigned int src_nents,
		unsigned long flags)
{
	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);
	struct dma_slave_config *dma_slave = chan->dma_slave;
	struct xgene_pcie_dma_desc *desc = NULL;
	struct scatterlist  *sg;
	int i, src_location, dst_location, slen = 0, glen = 0;
	u8 eop_intr = 0;

	/* basic sanity checks */
	if (dst_nents == 0 || src_nents == 0 || !dst_sg || !src_sg) {
		dev_err(chan->dev, "Incorrect input parameters\n");
		return NULL;
	}

	for_each_sg(src_sg, sg, src_nents, i)
		slen += sg_dma_len(sg);	

	for_each_sg(dst_sg, sg, dst_nents, i)
		glen += sg_dma_len(sg);

	if (slen != glen) {
		dev_err(chan->dev, "Scatter gather lenth mismatch %d %d\n", slen, glen);
		return NULL;
	}

	BUG_ON(!dma_slave);

	dev_dbg(chan->dev, "Channel%d: Prep slave sg\n",chan->id);

	if ((dma_slave->direction != DMA_MEM_TO_DEV) &&
			(dma_slave->direction != DMA_DEV_TO_MEM)) {
		dev_err(chan->dev, "Incorrect DMA Transfer direction %d\n", dma_slave->direction);
		return NULL;
	}

	desc = xgene_desc_get(chan);
	if(!desc) {
		dev_err(chan->dev, "Failed to get desc\n");
		return NULL;
	}

	desc->len = slen;
	desc->flags = flags;
	desc->txd.flags = flags;
	spin_lock_bh(&chan->lock);
	xgene_dma_get_next_queues(chan, desc);
	xgene_dma_update_sts_cnts(chan);

	/* Setup the in memory source, destination and status descriptors */
	src_location = (dma_slave->direction == DMA_MEM_TO_DEV) ? AXI_MEM : PCI_MEM;
	dst_location = (dma_slave->direction == DMA_MEM_TO_DEV) ? PCI_MEM : AXI_MEM;
	for_each_sg(src_sg, sg, src_nents, i) {
		if (i == (src_nents - 1))
			eop_intr = 1;
		xgene_dma_setup_src_q_desc(desc->src_desc,
				sg_dma_address(sg), sg_dma_len(sg), eop_intr,
				eop_intr, src_location);
		xgene_dma_update_src_cnts(chan);
		desc->src_desc++;
		dev_dbg (chan->dev, "%d %d %d %d %x %x %d\n",
				i, src_nents, eop_intr, src_location,
				upper_32_bits(sg_dma_address(sg)),
				lower_32_bits(sg_dma_address(sg)),
				sg_dma_len(sg));
	}
	for_each_sg(dst_sg, sg, dst_nents, i) {
			xgene_dma_setup_dst_q_desc(desc->dst_desc,
				sg_dma_address(sg), sg_dma_len(sg), dst_location);
			xgene_dma_update_dst_cnts(chan);
			desc->dst_desc++;
			dev_dbg (chan->dev, "%d %d %d %x %x %d",
					i, dst_nents, dst_location,
					upper_32_bits(sg_dma_address(sg)),
					lower_32_bits(sg_dma_address(sg)),
					sg_dma_len(sg));
	}
	spin_unlock_bh(&chan->lock);

	return &desc->txd;
}
/* xgene_pcie_dma_prep_slave_sg - Prep slave sg txn
 * @chan: chan for DMA transfer
 * @sgl: scatter gather list
 * @sg_len: length of sg txn
 * @direction: DMA transfer dirtn
 * @flags: DMA flags
 * @context: transfer context (ignored)
 *
 * Prepares sgl based periphral transfer
 */
static struct dma_async_tx_descriptor *xgene_pcie_dma_prep_slave_sg(
		struct dma_chan *dchan, struct scatterlist *sgl,
		unsigned int sg_len, enum dma_transfer_direction dir,
		unsigned long flags, void *context)
{

	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);
	struct dma_slave_config *dma_slave = chan->dma_slave;
	struct xgene_pcie_dma_desc *desc = NULL;
	struct scatterlist  *sg;
	int i;
	u8 eop_intr = 0;

	BUG_ON(!dma_slave);

	dev_dbg(chan->dev, "Channel%d: Prep slave sg\n",chan->id);

	if ((dir != DMA_MEM_TO_DEV) && (dir != DMA_DEV_TO_MEM)) {
		dev_err(chan->dev, "Incorrect DMA Transfer direction %d\n", dir);
		return NULL;
	}

	desc = xgene_desc_get(chan);
	if(!desc) {
		dev_err(chan->dev, "Failed to get desc\n");
		return NULL;
	}

	spin_lock_bh(&chan->lock);
	xgene_dma_get_next_queues(chan, desc);
	xgene_dma_update_sts_cnts(chan);

	/* Setup the in memory source, destination and status descriptors */
	desc->len = 0;
	for_each_sg(sgl, sg, sg_len, i) {
		if (dir == DMA_MEM_TO_DEV) {
			if (i == (sg_len - 1))
				eop_intr = 1;
			xgene_dma_setup_src_q_desc(desc->src_desc,
				sg_dma_address(sg), sg_dma_len(sg), eop_intr,
				eop_intr, AXI_MEM);
			xgene_dma_update_src_cnts(chan);
			desc->len += sg_dma_len(sg);
			desc->src_desc++;
		} else {
			xgene_dma_setup_dst_q_desc(desc->dst_desc,
				sg_dma_address(sg), sg_dma_len(sg), AXI_MEM);
			xgene_dma_update_dst_cnts(chan);
			desc->len += sg_dma_len(sg);
			desc->dst_desc++;
		}
	}

	if (dir == DMA_MEM_TO_DEV) {
		xgene_dma_setup_dst_q_desc(desc->dst_desc, dma_slave->dst_addr,
				desc->len, PCI_MEM);
		xgene_dma_update_dst_cnts(chan);
	} else {
		xgene_dma_setup_src_q_desc(desc->src_desc, dma_slave->src_addr,
				desc->len, 1, 1, PCI_MEM);
		xgene_dma_update_src_cnts(chan);
	}
	spin_unlock_bh(&chan->lock);

	return &desc->txd;
}


/**
 * xgene_pcie_dma_prep_memcpy -	Prep memcpy txn
 * @chan: chan for DMA transfer
 * @dest: destn address
 * @src: src address
 * @len: DMA transfer len
 * @flags: DMA flags
 *
 * Perform a DMA memcpy. Note we support slave periphral DMA transfers only
 * The periphral txn details should be filled in slave structure properly
 * Returns the descriptor for this txn
 * Right now we assume on AXI to AXI (may be mapped to PCIe) transfer.
 */
static struct dma_async_tx_descriptor *xgene_pcie_dma_prep_memcpy(
			struct dma_chan *dchan, dma_addr_t dest,
			dma_addr_t src, size_t len, unsigned long flags)
{

	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);
	struct xgene_pcie_dma_desc *desc = NULL;
	u8 src_location, dest_location;

	BUG_ON(!chan);

	dev_dbg(chan->dev, "Channel%d: Preparing for memcpy\n",chan->id);

	desc = xgene_desc_get(chan);
	if(!desc)
		goto err_desc_get;

	/* Setup the in memory source, destination and status descriptors */
	desc->len = len;
	desc->flags = flags;
	desc->txd.flags = flags;

	src_location  = (flags & DMA_COMPL_SRC_UNMAP_SINGLE) ? AXI_MEM : PCI_MEM;
	dest_location = (flags & DMA_COMPL_SRC_UNMAP_SINGLE) ? PCI_MEM : AXI_MEM;

	spin_lock_bh(&chan->lock);
	xgene_dma_get_next_queues(chan, desc);
	xgene_dma_update_src_cnts(chan);
	xgene_dma_update_dst_cnts(chan);
	xgene_dma_update_sts_cnts(chan);
	flush_to_ddr(desc->sts_desc, sizeof(struct sts_queue_desc));
	spin_unlock_bh(&chan->lock);

	xgene_dma_setup_src_q_desc(desc->src_desc, src, len, 1, 1, src_location);
	xgene_dma_setup_dst_q_desc(desc->dst_desc, dest, len, dest_location);

	return &desc->txd;
err_desc_get:
	dev_err(chan->dev, "Failed to get desc\n");
	xgene_desc_put(chan, desc);
	return NULL;
}

/**
 * xgene_pcie_dma_tx_status - Return status of txn
 * @chan: chan for where status needs to be checked
 * @cookie: cookie for txn
 * @txstate: DMA txn state
 *
 * Return status of DMA txn
 */
static enum dma_status xgene_pcie_dma_tx_status(struct dma_chan *dchan,
					      dma_cookie_t cookie,
					      struct dma_tx_state *txstate)
{
	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);
	dma_cookie_t last_used;
	dma_cookie_t last_complete;
	int ret;

	spin_lock_bh(&chan->lock);
	last_complete = chan->completed;
	last_used = dchan->cookie;

	ret = dma_async_is_complete(cookie, last_complete, last_used);
	if (ret != DMA_SUCCESS) {
		xgene_scan_descriptors(chan);
		last_complete = chan->completed;
		last_used = dchan->cookie;
		ret = dma_async_is_complete(cookie, last_complete, last_used);
	}

	if (txstate) {
		txstate->last = last_complete;
		txstate->used = last_used;
		txstate->residue = 0;
	}
	spin_unlock_bh(&chan->lock);
	return ret;
}

/**
 * xgene_pcie_dma_issue_pending -	callback to issue pending txn
 * @chan: chan where pending trascation needs to be checked and submitted
 *
 * Call for scan to issue pending descriptors
 */
static void xgene_pcie_dma_issue_pending(struct dma_chan *dchan)
{
	struct xgene_dma_chan *chan = to_xgene_pcie_dma_chan(dchan);

	xgene_start_dma(chan);
#if 0
	spin_lock_bh(&chan->lock);
	if (!list_empty(&chan->queue)) {
		list_splice_tail(&chan->queue, &chan->active_list);
		xgene_scan_descriptors(chan);
	}
	spin_unlock_bh(&chan->lock);
#endif
}

static void xgene_setup_dma_caps(struct xgene_dma_port *port)
{
	dma_cap_zero(port->common.cap_mask);
	dma_cap_set(DMA_SLAVE, port->common.cap_mask);
	dma_cap_set(DMA_MEMCPY, port->common.cap_mask);
	dma_cap_set(DMA_PRIVATE, port->common.cap_mask);

	port->common.device_alloc_chan_resources =
					xgene_pcie_dma_alloc_chan_resources;
	port->common.device_free_chan_resources =
					xgene_pcie_dma_free_chan_resources;
	port->common.device_prep_dma_memcpy = xgene_pcie_dma_prep_memcpy;
	port->common.device_tx_status = xgene_pcie_dma_tx_status;
	port->common.device_issue_pending = xgene_pcie_dma_issue_pending;
	port->common.device_prep_slave_sg = xgene_pcie_dma_prep_slave_sg;
	port->common.device_prep_dma_sg = xgene_pcie_dma_prep_sg;
	port->common.device_control = xgene_pcie_dma_device_control;
}
#ifdef CONFIG_XGENE_PCIE_DMA_POLLING
/* Implements polling mechanism to check if DMA is comeplete */
static int wait_for_dma_completion(struct xgene_dma_chan *chan,
				   struct xgene_pcie_dma_desc *desc)
{
	u32 status;
	int timeout = 1000;
	
	BUG_ON(chan->state == CHANNEL_DISABLED);

	do {
		xgene_dma_read(chan->base + DMA_CONTROL,
				&status);
		if(status & AXI_INTERRUPT_STATUS_MASK) {
			dev_dbg(chan->dev, "chan%d: XGENE DMA AXI Interrupt\n",
					chan->id);
			/* set bit again to clear interrupt */
			status |= AXI_INTERRUPT_STATUS_MASK;
			break;
		}
	} while(timeout--);

	if (timeout == -1)
		dev_dbg(chan->dev, "Request Timeout!\n");
	xgene_descriptor_complete(chan, desc);
	return 0;
}
#endif
/**
 * ISR and Tasklet
 */
static irqreturn_t xgene_dma_isr(int irq, void *data)
{
	struct xgene_dma_port *port = (struct xgene_dma_port *) data;
	struct xgene_dma_chan *chan;
	int index;
	u32 imask, status;

	for (index = 0; index < port->num_chan; index++) {
		chan = &port->chan[index];
		BUG_ON(chan->state == CHANNEL_DISABLED);
		xgene_dma_read(chan->base + DMA_CONTROL,
			     &status);
		dev_dbg(chan->dev, "Status : 0x%08x\n", status);
		if(status & AXI_INTERRUPT_STATUS_MASK) {
			imask = xgene_pcie_dma_read_intr(chan->port_num);
			imask |= (1 << 8);
			xgene_pcie_dma_write_intr(chan->port_num, imask);
			dev_dbg(chan->dev, "chan%d: XGENE DMA AXI Interrupt\n",
				   chan->id);
			/* set bit again to clear interrupt */
			status |= AXI_INTERRUPT_STATUS_MASK; 
			if(status & AXI_DMA_ERROR_MASK) {
				status |= AXI_DMA_ERROR_MASK; 
				dev_err(chan->dev, "chan%d: XGENE DMA AXI ERROR\n",
				   chan->id);
			}
			xgene_dma_write(chan->base + DMA_CONTROL, status);
			chan->irq_cnt++;
		} else if (status & PCIE_INTERRUPT_STATUS_MASK) {
			/* XXX: Not expected */
			dev_dbg(chan->dev, "chan%d: XGENE DMA PCIe Interrupt\n",
				   chan->id);
			status |= PCIE_INTERRUPT_STATUS_MASK;
		} else {
			dev_dbg(chan->dev, "chan%d: Spurious Interrupt\n", chan->id);
		}
	}

	dev_dbg(chan->dev, "Scheduling tasklet\n");
	/* Lets do further processing in bottom half */
	tasklet_schedule(&port->completion_tasklet);
	return IRQ_HANDLED;
}

/**
 * xgene_dma_tasklet -	DMA interrupt tasklet
 * @data: tasklet arg (the controller structure)
 *
 * Scan the controller for interrupts for completion/error
 * Clear the interrupt and call for handling completion/error
 */
static void xgene_dma_tasklet(unsigned long data)
{
	struct xgene_dma_port *port = (struct xgene_dma_port *) data;
	struct xgene_dma_chan *chan;
	int index;

	for (index = 0; index < port->num_chan; index++) {
		chan = &port->chan[index];
		xgene_scan_descriptors(chan);
		chan->tasklet_cnt++;
	}
}

/**
 * xgene_dma_alloc_sglists - Allocate Source, Destination and Status
 *			   descriptors sg lists
 * @chan: XGENE PCIe DMA channel structure
 *
 * Allocate the the DMA channel  Source, Destination and Status
 * descriptors sg lists
 */
static int xgene_dma_alloc_sglists(struct xgene_dma_chan *chan)
{
	unsigned long src_alloc_len;
	unsigned long dst_alloc_len;
	unsigned long sts_alloc_len;

	dev_dbg (chan->dev, "Size of src_queue_desc = %lu, dst_queue_desc = %lu,"
		  "sts_queue_desc = %lu\n", sizeof(struct src_queue_desc),
		  sizeof(struct dst_queue_desc), sizeof(struct sts_queue_desc));
	src_alloc_len = chan->src_elems * sizeof(struct src_queue_desc);
	chan->src_elem_addr = dmam_alloc_coherent(chan->dev, src_alloc_len,
			&chan->src_elem_addr_phys, GFP_ATOMIC);
	if(!chan->src_elem_addr) {
		dev_err(chan->dev, "Failed to allocate source sg descriptors\n");
		return -ENOMEM;
	}
	dev_dbg (chan->dev, "src_elem_addr: 0x%p src_elem_addr_phys: 0x%08llx "
		  "src_alloc_len : %ld\n", chan->src_elem_addr,
		  (unsigned long long)chan->src_elem_addr_phys,
		  src_alloc_len);
	memset(chan->src_elem_addr, 0, src_alloc_len);

	dst_alloc_len = chan->dst_elems * sizeof(struct dst_queue_desc);
	chan->dst_elem_addr = dmam_alloc_coherent(chan->dev, dst_alloc_len,
			&chan->dst_elem_addr_phys, GFP_ATOMIC);
	if(!chan->dst_elem_addr) {
		 dev_err(chan->dev, "Failed to allocate destination sg descriptors\n");
		return -ENOMEM;
	}
	dev_dbg (chan->dev, "dst_elem_addr: 0x%p dst_elem_addr_phys: 0x%08llx "
		  "dst_alloc_len : %ld\n", chan->dst_elem_addr,
		   (unsigned long long)chan->dst_elem_addr_phys,
		   dst_alloc_len);
	memset(chan->dst_elem_addr, 0, dst_alloc_len);

	sts_alloc_len = chan->sts_elems * sizeof(struct sts_queue_desc);
	chan->sts_elem_addr = dmam_alloc_coherent(chan->dev, sts_alloc_len,
			&chan->sts_elem_addr_phys, GFP_ATOMIC);
	if(!chan->sts_elem_addr) {
		dev_err (chan->dev, "Failed to allocate source sg descriptors\n");
		return -ENOMEM;
	}
	dev_dbg (chan->dev, "sts_elem_addr: 0x%p sts_elem_addr_phys: 0x%08llx "
		  "sts_alloc_len : %ld\n", chan->sts_elem_addr,
		   (unsigned long long)chan->sts_elem_addr_phys,
		   sts_alloc_len);
	memset(chan->sts_elem_addr, 0, sts_alloc_len);
	return 0;
}

/**
 * xgene_init_dma_channel - Initialize the DMA controller HW
 * @chan: XGENE PCIe DMA channel structure
 *
 * Reset DMA channel. Initialize the DMA channel source,
 * destination and status queues. Enable DMA channel.
 */

static int xgene_init_dma_channel(struct xgene_dma_chan *chan)
{
	/* XXX: Should we bail out here? */
	/* Disable DMA Channel fiels */
	disable_axi_dma_interrupts(chan);
	disable_pci_dma_interrupts(chan);
	disable_dma_channel(chan);
	reset_dma_channel(chan);
	/* Now setup queue management registers */
	xgene_dma_write(chan->base + SRC_Q_NEXT, 0);
	xgene_dma_write(chan->base + DST_Q_NEXT, 0);
	xgene_dma_write(chan->base + STA_Q_NEXT, 0);

	xgene_dma_write(chan->base + SRC_Q_PTR_LO,
			LODWORD(chan->src_elem_addr_phys) | AXI_MEM |
			AXI_DESC_COHERENCY);
	xgene_dma_write(chan->base + SRC_Q_PTR_HI,
			HIDWORD(chan->src_elem_addr_phys));
	xgene_dma_write(chan->base + SRC_Q_SIZE, chan->src_elems);

	xgene_dma_write(chan->base + DST_Q_PTR_LO,
			LODWORD(chan->dst_elem_addr_phys) | AXI_MEM |
			AXI_DESC_COHERENCY);
	xgene_dma_write(chan->base + DST_Q_PTR_HI,
			HIDWORD(chan->dst_elem_addr_phys));
	xgene_dma_write(chan->base + DST_Q_SIZE, chan->dst_elems);

	xgene_dma_write(chan->base + STA_Q_PTR_LO,
			LODWORD(chan->sts_elem_addr_phys) | AXI_MEM |
			AXI_DESC_COHERENCY);
	xgene_dma_write(chan->base + STA_Q_PTR_HI,
			HIDWORD(chan->sts_elem_addr_phys));
	xgene_dma_write(chan->base + STA_Q_SIZE, chan->sts_elems);
	xgene_dma_write(chan->base + SRC_Q_LIMIT, 0);
	xgene_dma_write(chan->base + DST_Q_LIMIT, 0);
	xgene_dma_write(chan->base + STA_Q_LIMIT, 0);
	chan->state = CHANNEL_READY;
	enable_dma_channel(chan);
	return 0;
}

/**
 * xgene_setup_dma - Setup the DMA controller
 * @info: Driver probe information structure
 *
 * Initialize the DMA controller, channels, registers with DMA engine,
 * ISR. Initialize DMA controller channels.
 */
static int xgene_setup_dma_channel(struct xgene_dma_port *port,
		struct xgene_pcie_info *info,
		void __iomem * dma_sync,
		int port_num)
{
	struct xgene_dma_chan *chan;
	int i;
	int ret = 0;

	INIT_LIST_HEAD(&port->common.channels);
	for (i = 0; i < port->num_chan; i++) {
		chan = &port->chan[i];
		memset(chan, 0, sizeof(*chan));
		chan->state = CHANNEL_DISABLED;
		chan->id = i;
		chan->dev = port->common.dev;
		chan->base = port->dma_reg + (i * CHAN_REG_SPACE);
		chan->chan.device = &port->common;
		chan->chan.cookie =  1;
		chan->dma_sync = dma_sync;
		chan->port_num = port_num;
		if(!is_dma_channel_present(chan)) {
			 dev_err(port->common.dev, "DMA Channel%d is disabled\n",chan->id);
			continue;
		}

		dev_dbg(port->common.dev, "Found DMA Channel %d\n", i);
		chan->src_elems = info->src_elems;
		chan->dst_elems = info->dst_elems;
		chan->sts_elems = info->sts_elems;
		dev_dbg(port->common.dev,
				"src_elems : %u dest_elems : %u sts_elems : %u\n",
				chan->src_elems, chan->dst_elems,
				chan->sts_elems);

		ret = xgene_dma_alloc_sglists(chan);
		if(ret)
			continue;
		INIT_LIST_HEAD(&chan->queue);
		INIT_LIST_HEAD(&chan->descs);
		spin_lock_init(&chan->lock);
		list_add_tail(&chan->chan.device_node,
			      &port->common.channels);
	}

	return 0;
}

static void xgene_adjust_dma_attributes(struct xgene_pcie_info *info)
{
	info->max_chan = 4;
	info->src_elems = MAX_SRC_SG_LISTS;
	info->dst_elems = MAX_DST_SG_LISTS;
	info->sts_elems = MAX_STS_SG_LISTS;
}

/**
 * xgene_pcie_dma_probe - PCI Probe
 * @id: pci device id structure
 *
 * Initialize the DMA port and DMA channels.
 * Call setup_dma to complete contoller and chan initilzation
 */
void xgene_pciedma_hw_init(void)
{
	int i;
	struct xgene_dma_port *port = port_g;
	struct xgene_dma_chan *chan;

	for (i = 0; i < port->num_chan; i++) {
		chan = &port->chan[i];
		xgene_init_dma_channel(chan);
	}
}
EXPORT_SYMBOL(xgene_pciedma_hw_init);

static int xgene_pciedma_probe (struct platform_device *pdev)
{
	int err, port_num;
	struct resource *res;
	void __iomem *dma_sync;
	struct xgene_pcie_info info;
	static struct xgene_dma_port *port;
	struct device_node *np = of_node_get(pdev->dev.of_node);

	port_g = port = devm_kzalloc(&pdev->dev, sizeof(*port), GFP_KERNEL);
	if (!port) {
		dev_err(&pdev->dev, "kzalloc failed XGENE PCIe DMA probe\n");
		return -ENOMEM;
	}

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "ama-addr");
	dma_sync = ioremap_nocache(res->start, (res->end - res->start));
	if (IS_ERR(dma_sync)) {
		dev_err (&pdev->dev, "Could not map ama-addr address\n");
		return -ENOMEM;
	}
	if (of_property_read_u32(np, "port_num", &port_num))
		port_num = 0;
	port->dma_reg = xgene_pcie_dma_cfg_base(port_num);
	if (port->dma_reg == NULL) {
		dev_err (&pdev->dev, "Could not get dma-cfg-base address\n");
		return -ENOMEM;
	}
	pdev->dev.dma_mask = &pdev->dev.coherent_dma_mask;
	pdev->dev.coherent_dma_mask = DMA_BIT_MASK(64);
	xgene_adjust_dma_attributes(&info);

	port->common.dev = &pdev->dev;
	port->num_chan = info.max_chan;
	err = xgene_setup_dma_channel(port, &info, dma_sync, port_num);
	if (err) {
		dev_err (&pdev->dev, "setup channel failed\n");
		return -EINVAL;
	}

	if (!enable_polling) {
		res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
		if (!res) {
			dev_err(&pdev->dev, "No irq resource\n");
			return -EINVAL;
		}

		tasklet_init(&port->completion_tasklet, xgene_dma_tasklet,
				(long unsigned int)port);

		err = devm_request_irq(&pdev->dev, res->start, xgene_dma_isr,
				IRQF_SHARED, "PCIEDMA", port);
		dev_dbg(&pdev->dev, "Requesting IRQ %d for XGENE PCIe DMA\n", (int)res->start);
		if (err) {
			dev_err(&pdev->dev, "Request IRQ failed for XGENE PCIe DMA\n");
			return -EINVAL;
		}

	}
	xgene_setup_dma_caps(port);
	err = dma_async_device_register(&port->common);
	if (err) {
		dev_err(port->common.dev, "XGENE PCIe DMA device_register failed: %d\n", err);
		return -EINVAL;
	}

	INIT_LIST_HEAD(&dma_port_list);
	list_add_tail(&port->list, &dma_port_list);
	platform_set_drvdata(pdev, port);
	dev_info (&pdev->dev, "X-Gene PCIE DMA driver probed\n");
	return 0;
}

static int xgene_pciedma_remove(struct platform_device *pdev)
{
	struct xgene_dma_port *port = platform_get_drvdata(pdev);
	struct xgene_dma_port *p1;

	list_for_each_entry(p1, &dma_port_list, list) {
		if(p1 == port) {
			list_del(&p1->list);
			dma_async_device_unregister(&p1->common);
			return 0;
		}
	}
	return -ENODEV;
}

static const struct of_device_id xgene_pciedma_match_table[] = {
	{.compatible = "xgene,xgene-pciedma",},
	{},
};

static struct platform_driver xgene_pciedma_driver = {
	.driver = {
		   .name = "xgene_pciedma",
		   .owner = THIS_MODULE,
		   .of_match_table = of_match_ptr(xgene_pciedma_match_table),
	},
	.probe = xgene_pciedma_probe,
	.remove = xgene_pciedma_remove,
};
module_platform_driver(xgene_pciedma_driver);

MODULE_AUTHOR("Applied Micro Circuit Corporation");
MODULE_DESCRIPTION("XGENE X-Gene PCIe DMA Driver");
MODULE_LICENSE("GPL v2");
