/**
 * APM X-Gene PCIe Driver
 *
 * Copyright (c) 2013 Applied Micro Circuits Corporation.
 *
 * Author: Tanmay Inamdar <tinamdar@apm.com>, Mayuresh Chitale <mchitale@apm.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/cacheflush.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/of_device.h>
#include <linux/dmaengine.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/mm.h>
#include <linux/random.h>
#include <linux/dmaengine.h>
#include <linux/dma/xgene-pcie-dmalib.h>
#define XGENE_DMA_MAX_LEN	0x400000 
static struct xgene_dma_completion cmpl_param;
static atomic_t compl_count;

int xgene_dma_get_compl_count(void)
{
	return (atomic_read(&compl_count));
}
EXPORT_SYMBOL(xgene_dma_get_compl_count);
void xgene_dma_set_compl_count(int val)
{
	atomic_set(&compl_count, val);
}
EXPORT_SYMBOL(xgene_dma_set_compl_count);
static void cmpl_callback(void *param)
{
	struct xgene_dma_completion *p = param;
	struct scatterlist *sg;
	int i;

	if (p->sgl) {
		for_each_sg(p->sgl, sg, p->sg_cnt, i) {
			dev_dbg (p->device->dev, "Unmapping %llX size %X\n",
					sg_dma_address(sg), sg_dma_len(sg));	
			dma_unmap_page(p->device->dev, sg_dma_address(sg),
					sg_dma_len(sg), p->slave->direction);
		}
		kfree(p->sgl);
		p->sgl = NULL;
	}
	if (p->slave) {
		kfree(p->slave);
		p->slave = NULL;
	}
	atomic_set(&p->flag, 1);
	wake_up(&p->cmpl);
	atomic_inc(&compl_count);
	dev_dbg(p->device->dev, "SG DMA Done %d\n", atomic_read(&compl_count));
}

static bool test_dma_filter_fn(struct dma_chan *chan, void *chan_id)
{
	dev_dbg(chan->device->dev, "%s: Chan ID: %d , Param ID: %d\n",
		__func__, chan->chan_id, *(u32 *)chan_id);
	return (chan->chan_id == *(u32 *)chan_id);
}

struct dma_chan *xgene_pcie_dma_request_channel (int chan)
{
	dma_cap_mask_t mask;
	int dma_cap = DMA_SLAVE;

	dma_cap_zero(mask);
	dma_cap_set(dma_cap, mask);
	dma_cap_set(DMA_PRIVATE, mask);
	return dma_request_channel(mask, test_dma_filter_fn, &chan);
}
EXPORT_SYMBOL(xgene_pcie_dma_request_channel);

struct dma_addr_map *xgene_pcie_dma_alloc_mem(struct device *dev, int rand,
		enum dma_transfer_direction dir, dma_addr_t ep_addr)
{
	u32 len = XGENE_DMA_MAX_LEN, seglen;
	int i = 0, count;
	struct dma_addr_map *map;

	switch (rand) {
		case 0:
			count = 1;
			break;
		case 1:
			get_random_bytes(&i, 2);
			count = XGENE_DMA_MAX_LEN / i;
			count = (count > 32) ? 32 : count;
			if (!count)
				count = 1;
			break;
		default:
			count = rand;
			len *= rand;
			dev_dbg (dev, "Allocating %X bytes of memory\n", len);
			break;
	}
	dev_info (dev, "Creating sglist with %d elems\n", count);
	map = kzalloc((count + 1) * sizeof(struct dma_addr_map), GFP_KERNEL);
	if (!map) {
		dev_err (dev, "Could not allocate map\n");
		return NULL;
	}
	seglen = len/count;
	for (i = 0; i < count; i++) {
		if (i == (count -1))
			seglen = len;
		else {
			len -= seglen;
		}
		map[i].len = seglen;
		if (ep_addr) {
			map[i].phys = ep_addr;
			ep_addr += seglen;
		} else {
			map[i].virt = kmalloc(seglen, GFP_KERNEL);
			get_random_bytes(map[i].virt, seglen);
			map[i].phys = dma_map_single(dev, (void *)map[i].virt,
					seglen, dir);
		}
	}
	return map;
}
EXPORT_SYMBOL(xgene_pcie_dma_alloc_mem);

struct scatterlist *xgene_pcie_dma_create_sg(struct dma_addr_map *map)
{
	int i =0, count = 0;
	struct scatterlist *sg = NULL, *sglist = NULL;

	while(map[count].len) count++;
	sglist = kzalloc(count  * sizeof(struct scatterlist), GFP_KERNEL);
	if (!sglist) {
		return sglist;
	}
	sg_init_table(sglist, count);
	for (i = 0, sg = sglist; i < count; i++) {
		sg_dma_address(sg) = map[i].phys;
		sg_dma_len(sg) = map[i].len;
		if (i != (count - 1))
			sg = sg_next(sg);
	}
	sg_mark_end(sg);
	return sglist;
}

int xgene_pcie_dma_sg(struct dma_chan *chan, struct scatterlist *src, struct scatterlist *dst, int dir)
{
	struct dma_device *dma_dev = chan->device;
	struct device *dev = dma_dev->dev;
	int flags = 0, dst_nents = 0, src_nents = 0;
	dma_cookie_t cookie;
	struct scatterlist *sg = NULL;
	struct dma_slave_config *dma_slave;
	struct dma_async_tx_descriptor *tx = NULL;

	BUG_ON(chan == NULL);
	BUG_ON(src == NULL);
	BUG_ON(dst == NULL);

	sg = src;	
	while (sg) {
		src_nents++;	
		sg = sg_next(sg);
	}
	sg = dst;	
	while (sg) {
		dst_nents++;	
		sg = sg_next(sg);
	}
	dev_dbg (dev, "src_ents %d dest_ents %d\n",src_nents, dst_nents);
	dma_slave = kzalloc(sizeof(*dma_slave), GFP_KERNEL);
	if (!dma_slave) {
		dev_err(dev, "Failed to allocate dma_slave\n");
		return -ENOMEM;
	}
	init_waitqueue_head(&cmpl_param.cmpl);
	atomic_set(&cmpl_param.flag, 0);
	dma_slave->direction = dir;
	flags = DMA_CTRL_ACK | DMA_PREP_INTERRUPT;
	/* device_control Required set direction in driver */
	dma_dev->device_control(chan, DMA_SLAVE_CONFIG,
			(unsigned long) dma_slave);

	tx = dma_dev->device_prep_dma_sg(chan, dst, dst_nents,
			src, src_nents, 
			flags);
	BUG_ON(tx == NULL);
	tx->callback = cmpl_callback;
	tx->callback_param = &cmpl_param;
	cmpl_param.slave = dma_slave;
	cmpl_param.device = dma_dev;
	if (dir == DMA_DEV_TO_MEM) {
		cmpl_param.sgl = dst;
		cmpl_param.sg_cnt = dst_nents;
	}
	if (dir == DMA_MEM_TO_DEV) {
		cmpl_param.sgl = src;
		cmpl_param.sg_cnt = src_nents;
	}
	cookie = tx->tx_submit(tx);
#if 0
	if (!wait_event_timeout(cmpl_param.cmpl,
				atomic_read(&cmpl_param.flag),100)) {
		pr_err("wait_event timeout - 1\n");
		return -EAGAIN;
	}
#endif
	return 0;
}

struct dma_addr_map *xgene_pcie_dma_sg_write(struct dma_chan *chan, dma_addr_t pci_addr, int rand)
{
	struct device *dev = chan->device->dev;
	struct scatterlist *src = NULL, *dst = NULL;
	struct dma_addr_map *smap = NULL, *dmap = NULL;

	dev_info (dev, "DMA Write\n");
	smap = xgene_pcie_dma_alloc_mem(dev, rand, DMA_MEM_TO_DEV, 0);
	src = xgene_pcie_dma_create_sg(smap);
	if (!src)
		return NULL;

	dmap = xgene_pcie_dma_alloc_mem(dev, rand, DMA_MEM_TO_DEV, pci_addr);
	dst = xgene_pcie_dma_create_sg(dmap);
	
	if (xgene_pcie_dma_sg(chan, src, dst, DMA_MEM_TO_DEV)) {
		kfree(smap);
		smap = NULL;
	}
	kfree(dmap);
	return  smap;
}

struct dma_addr_map *xgene_pcie_dma_sg_read(struct dma_chan *chan, dma_addr_t pci_addr, int rand)
{
	struct device *dev = chan->device->dev;
	struct scatterlist *src = NULL, *dst = NULL;
	struct dma_addr_map *smap = NULL, *dmap = NULL;

	dev_info (dev, "DMA Read\n");
	dmap = xgene_pcie_dma_alloc_mem(dev, rand, DMA_DEV_TO_MEM, 0);
	dst = xgene_pcie_dma_create_sg(dmap);
	if (!dst)
		return NULL;

	smap = xgene_pcie_dma_alloc_mem(dev, rand, DMA_DEV_TO_MEM, pci_addr);
	src = xgene_pcie_dma_create_sg(smap);
	
	if (xgene_pcie_dma_sg(chan, src, dst, DMA_DEV_TO_MEM)) {
		kfree(dmap);
		dmap = NULL;
	}
	kfree(smap);
	return  dmap;
}

int xgene_pcie_dma_memcpy(struct dma_chan *chan, u64 dest, u64 src,
				 size_t len, enum dma_transfer_direction dir)
{
	struct dma_device *dev = chan->device;
	struct dma_async_tx_descriptor *tx;
	dma_addr_t dma_dest, dma_src;
	dma_cookie_t cookie;
	unsigned long flags;

	init_waitqueue_head(&cmpl_param.cmpl);
	atomic_set(&cmpl_param.flag, 0);

	if (dir == DMA_MEM_TO_DEV) {
		dma_src = dma_map_single(dev->dev, (void *)src, len, DMA_MEM_TO_DEV);
		dma_dest = (dma_addr_t) dest;
		flags = DMA_COMPL_SRC_UNMAP_SINGLE;
	} else {
		dma_src = (dma_addr_t) src;
		dma_dest = dma_map_single(dev->dev, (void *)dest, len,
					  DMA_DEV_TO_MEM);
		flags = DMA_COMPL_DEST_UNMAP_SINGLE;
	}

	flags |= DMA_CTRL_ACK | DMA_PREP_INTERRUPT;
	tx = dev->device_prep_dma_memcpy(chan, dma_dest, dma_src, len, flags);
	if (!tx) {
		pr_err ("%s: Device prep memcpy failed\n", __func__);
		if (flags & DMA_COMPL_SRC_UNMAP_SINGLE)
			dma_unmap_single(dev->dev, dma_src, len, DMA_MEM_TO_DEV);
		else
			dma_unmap_single(dev->dev, dma_dest, len, DMA_DEV_TO_MEM);
		return -ENOMEM;
	}
	tx->callback = cmpl_callback;
	tx->callback_param = &cmpl_param;
	cmpl_param.sgl = NULL;
	cmpl_param.slave = NULL;
	cmpl_param.device = dev;

	cookie = tx->tx_submit(tx);
#if 0
	preempt_disable();
	__this_cpu_add(chan->local->bytes_transferred, len);
	__this_cpu_inc(chan->local->memcpy_count);
	preempt_enable();
	if (!wait_event_timeout(cmpl_param.cmpl,
				atomic_read(&cmpl_param.flag),100)) {
		pr_err("%s: timeout\n", __func__);
		return -EAGAIN;
	}
#endif
	return 0;
}
EXPORT_SYMBOL(xgene_pcie_dma_memcpy);
