/*
 * Applied Micro X-Gene SoC DMA engine Driver
 *
 * Copyright (c) 2015, Applied Micro Circuits Corporation
 * Authors: Rameshwar Prasad Sahu <rsahu@apm.com>
 *	    Loc Ho <lho@apm.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/acpi.h>
#include <linux/bitrev.h>
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/dmaengine.h>
#include <linux/dmapool.h>
#include <linux/efi.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/of_device.h>
#include <linux/pm_runtime.h>
#include <misc/xgene/qmtm/xgene_qmtm.h>
#include <misc/xgene/xgene-pktdma.h>

#include "dmaengine.h"

/* DMA device csr registers and bit definitions */
#define DMA_IPBRR		0x0
#define DMA_DEV_ID_RD(v)	((v) & 0x00000FFF)
#define DMA_BUS_ID_RD(v)	(((v) >> 12) & 3)
#define DMA_REV_NO_RD(v)	(((v) >> 14) & 3)
#define DMA_GCR			0x10
#define DMA_CH_SETUP(v)		((v) = ((v) & ~0x000FFFFF) | 0x000AAFFF)
#define DMA_ENABLE(v)		((v) |= BIT(31))
#define DMA_DISABLE(v)		((v) &= ~BIT(31))
#define DMA_RAID6_CONT		0x14
#define DMA_RAID6_MULTI_CTRL(v)	((v) << 24)
#define DMA_INT			0x70
#define DMA_INT_MASK		0x74
#define DMA_INT_ALL_MASK	0xFFFFFFFF
#define DMA_INT_ALL_UNMASK	0x0
#define DMA_INT_MASK_SHIFT	0x14
#define DMA_RING_INT0_MASK	0x90A0
#define DMA_RING_INT1_MASK	0x90A8
#define DMA_RING_INT2_MASK	0x90B0
#define DMA_RING_INT3_MASK	0x90B8
#define DMA_RING_INT4_MASK	0x90C0
#define DMA_CFG_RING_FQ_ASSOC	0x90DC
#define DMA_CFG_RING_WQ_ASSOC	0x90E0
#define DMA_ASSOC_RING_MNGR1	0xFFFFFFFF
#define DMA_MEM_RAM_SHUTDOWN	0xD070
#define DMA_BLK_MEM_RDY		0xD074
#define DMA_BLK_MEM_RDY_VAL	0xFFFFFFFF

/* SoC EFUSE csr register and bit defination */
#define SOC_JTAG1_SHADOW	0x18
#define PQ_DISABLE_MASK		BIT(13)

/* DMA Descriptor format */
#define BIT_ULL(nr)			(1ULL << (nr))
#define DMA_DESC_ELERR_POS		46
#define DMA_DESC_NV_BIT			BIT_ULL(50)
#define DMA_DESC_IN_BIT			BIT_ULL(55)
#define DMA_DESC_RTYPE_POS		56
#define DMA_DESC_LERR_POS		60
#define DMA_DESC_BUFLEN_POS		48
#define DMA_DESC_C_BIT			BIT_ULL(63)
#define DMA_DESC_DR_BIT			BIT_ULL(61)
#define DMA_DESC_HOENQ_POS		48
#define DMA_DESC_FLYBY_POS		4
#define DMA_DESC_BD_BIT			BIT(0)
#define DMA_DESC_SD_BIT			BIT(1)
#define DMA_DESC_CRCSEED_POS		8
#define	DMA_DESC_FPQ_NUM_POS		32
#define DMA_RING_DESC_CNT_POS		1
#define DMA_RING_DESC_CNT_MASK		0x0001FFFE
#define DMA_DESC_STATUS(elerr, lerr)	(((elerr) << 4) | (lerr))
#define DMA_DESC_CRC_RESULT_RD(m)	(u32)((u64 *)(m))[3]
#define DMA_DESC_ELERR_RD(m)		\
	((((u64 *)(m))[0] >> DMA_DESC_ELERR_POS) & 0x3)
#define DMA_DESC_NV_SET(m)		\
	(((u64 *)(m))[0] |= DMA_DESC_NV_BIT)
#define DMA_DESC_IN_SET(m)		\
	(((u64 *)(m))[0] |= DMA_DESC_IN_BIT)
#define DMA_DESC_RTYPE_SET(m, v)	\
	(((u64 *)(m))[0] |= ((u64)(v) << DMA_DESC_RTYPE_POS))
#define DMA_DESC_LERR_RD(m)		\
	((((u64 *)(m))[0] >> DMA_DESC_LERR_POS) & 0x7)
#define DMA_DESC_BUFADDR_SET(m, v)	\
	(((u64 *)(m))[0] |= (v))
#define DMA_DESC_BUFLEN_SET(m, v)	\
	(((u64 *)(m))[0] |= ((u64)(v) << DMA_DESC_BUFLEN_POS))
#define DMA_DESC_C_SET(m)		\
	(((u64 *)(m))[1] |= DMA_DESC_C_BIT)
#define DMA_DESC_DR_SET(m)		\
	(((u64 *)(m))[2] |= DMA_DESC_DR_BIT)
#define DMA_DESC_DST_ADDR_SET(m, v)	\
	(((u64 *)(m))[3] |= (v))
#define DMA_DESC_H0ENQ_NUM_SET(m, v)	\
	(((u64 *)(m))[3] |= ((u64)(v) << DMA_DESC_HOENQ_POS))
#define DMA_DESC_FLYBY_SET(m, v)	\
	(((u64 *)(m))[2] |= ((v) << DMA_DESC_FLYBY_POS))
#define DMA_DESC_MULTI_SET(m, v, i)	\
	(((u64 *)(m))[2] |= ((u64)(v) << (((i) + 1) * 8)))
#define DMA_DESC_NV_RD(m)		\
	(((u64 *)(m))[0] & DMA_DESC_NV_BIT)
#define DMA_DESC_BD_SET(m)		\
	(((u64 *)(m))[2] |= DMA_DESC_BD_BIT)
#define DMA_DESC_SD_SET(m)		\
	(((u64 *)(m))[2] |= DMA_DESC_SD_BIT)
#define DMA_DESC_CRCSEED_SET(m, v)	\
	(((u64 *)(m))[2] |= ((u64)(v) << DMA_DESC_CRCSEED_POS))
#define DMA_DESC_CRCBYTECNT_SET(m, v)	\
	(((u64 *)(m))[3] |= ((v) & 0x0000FFFF))
#define DMA_DESC_FPQ_NUM_SET(m, v)	\
	(((u64 *)(m))[0] |= ((u64)(v) << DMA_DESC_FPQ_NUM_POS))

/* DMA descriptor empty s/w signature */
#define DMA_DESC_SET_EMPTY(m)		\
	(((u32 *)(m))[EMPTY_SLOT_INDEX] = EMPTY_SLOT)
#define DMA_DESC_IS_EMPTY(m)		\
	(((u32 *)(m))[EMPTY_SLOT_INDEX] == EMPTY_SLOT)

/* Define macros for DMA property */
#define DMA_RING_OWNER			0x03
#define DMA_QMTM_IP			0x1
#define DMA_MAX_CHANNEL			4
#define DMA_XOR_CHANNEL			0
#define DMA_PQ_CHANNEL			1
#define DMA_COPY_CHANNEL		2
#define DMA_FLYBY_CHANNEL		3
#define DMA_MAX_BYTE_CNT		0x4000	/* 16 KB */
#define DMA_MAX_64B_DESC_BYTE_CNT	0x14000 /* 80 KB */
#define DMA_XOR_ALIGNMENT		6
#define DMA_MAX_XOR_SRC			5
#define DMA_16K_BUFFER_LEN_CODE		0x0
#define DMA_INVALID_LEN_CODE		0x7800

/* Define to enable DMA operations on channels */
#undef	DMA_ENABLE_MEMCPY
#define DMA_ENABLE_SG
#define DMA_ENABLE_XOR
#define DMA_ENABLE_PQ
#undef	DMA_ENABLE_MEMCPY_ALL_CHANNEL
#undef	DMA_ENABLE_SG_ALL_CHANNEL

/* DMA descriptor error codes */
#define ERR_DESC_AXI		0x01
#define ERR_BAD_DESC		0x02
#define ERR_READ_DATA_AXI	0x03
#define ERR_WRITE_DATA_AXI	0x04
#define ERR_FBP_TIMEOUT		0x05
#define ERR_ECC			0x06
#define ERR_DIFF_SIZE		0x08
#define ERR_SCT_GAT_LEN		0x09
#define ERR_CRC_ERR		0x11
#define ERR_CHKSUM		0x12
#define ERR_DIF			0x13

/* DMA error interrupt codes */
#define ERR_DIF_SIZE_INT	0x0
#define ERR_GS_ERR_INT		0x1
#define ERR_FPB_TIMEO_INT	0x2
#define ERR_WFIFO_OVF_INT	0x3
#define ERR_RFIFO_OVF_INT	0x4
#define ERR_WR_TIMEO_INT	0x5
#define ERR_RD_TIMEO_INT	0x6
#define ERR_WR_ERR_INT		0x7
#define ERR_RD_ERR_INT		0x8
#define ERR_BAD_DESC_INT	0x9
#define ERR_DESC_DST_INT	0xA
#define ERR_DESC_SRC_INT	0xB

/* DMA flyby operation code */
#define FLYBY_2SRC_XOR		0x8
#define FLYBY_3SRC_XOR		0x9
#define FLYBY_4SRC_XOR		0xA
#define FLYBY_5SRC_XOR		0xB

/* DMA SW descriptor flags */
#define DMA_FLAG_64B_DESC	BIT(0)
#define DMA_FLAG_SG_ACTIVE	BIT(1)
#define DMA_FLAG_FLYBY_ACTIVE	BIT(2)
#define DMA_FLAG_FIRST_DST	BIT(3)
#define DMA_FLAG_SECOND_DST	BIT(4)

#define to_dma_desc_sw(tx)	\
	container_of(tx, struct xgene_dma_desc_sw, tx)
#define to_dma_chan(dchan)	\
	container_of(dchan, struct xgene_dma_chan, dma_chan)

#define chan_dbg(chan, fmt, arg...)	\
	dev_dbg(chan->dev, "%s: " fmt, chan->name, ##arg)
#define chan_err(chan, fmt, arg...)	\
	dev_err(chan->dev, "%s: " fmt, chan->name, ##arg)

typedef void (*dma_flyby_cb)(void *flyby_ctx, u32 flyby_result);

struct xgene_dma_desc_sw {
	struct xgene_qmtm_msg32 desc1;
	struct xgene_qmtm_msg32 desc2;
	u32 flags;

	/* Shadow copy to use in unmap */
	size_t len;
	u32 src_cnt;
	dma_addr_t dst[2];
	dma_addr_t src[DMA_MAX_XOR_SRC];

	/* For flyby operations */
	struct scatterlist *src_sg;
	void *flyby_ctx;
	dma_flyby_cb flyby_cb;

	struct list_head node;
	struct list_head tx_list;
	struct dma_async_tx_descriptor tx;
};

/**
 * struct xgene_dma_chan - internal representation of an DMA channel
 * @dma_chan: dmaengine channel object member
 * @pdma: DMA device structure reference
 * @dev: struct device reference for dma mapping api
 * @id: raw id of this channel
 * @rx_irq: channel IRQ
 * @name: name of DMA channel
 * @lock: serializes enqueue/dequeue operations to the descriptor pool
 * @pending: number of transaction request pushed to DMA controller for
 *	execution, but still waiting for completion,
 * @max_outstanding: max number of outstanding request we can push to channel
 * @ld_pending: descriptors which are queued to run, but have not yet been
 *	submitted to the hardware for execution
 * @ld_running: descriptors which are currently being executing by the hardware
 * @ld_completed: descriptors which have finished execution by the hardware.
 *	These descriptors have already had their cleanup actions run. They
 *	are waiting for the ACK bit to be set by the async tx API.
 * @desc_pool: descriptor pool for DMA operations
 * @tasklet: bottom half where all completed descriptors cleans
 * @tx_ring: transmit ring descriptor that we use to prepare actual
 *	descriptors for further executions
 * @rx_ring: receive ring descriptor that we use to get completed DMA
 *	descriptors during cleanup time
 * @bufpool: Queue which maintains list of allocated memory for flyby operations
 *	needed by DMA engine
 */
struct xgene_dma_chan {
	struct dma_chan dma_chan;
	struct xgene_dma *pdma;
	struct device *dev;
	int id;
	int rx_irq;
	char name[8];
	spinlock_t lock;
	int pending;
	int max_outstanding;
	struct list_head ld_pending;
	struct list_head ld_running;
	struct list_head ld_completed;
	struct dma_pool *desc_pool;
	struct tasklet_struct tasklet;
	struct xgene_qmtm_qinfo tx_ring;
	struct xgene_qmtm_qinfo rx_ring;
	struct xgene_qmtm_qinfo bufpool;
};

/**
 * struct xgene_dma - internal representation of an DMA device
 * @err_irq: DMA error irq number
 * @csr_dma: base for DMA register access
 * @csr_efuse: base for efuse register access
 * @sdev : reference to queue manager used by DMA engine
 * @dma_dev: embedded struct dma_device
 * @chan: reference to DMA channels
 */
struct xgene_dma {
	struct device *dev;
	struct clk *clk;
	int err_irq;
	void __iomem *csr_dma;
	void __iomem *csr_efuse;
	struct xgene_qmtm_sdev *sdev;
	struct dma_device dma_dev[DMA_MAX_CHANNEL];
	struct xgene_dma_chan chan[DMA_MAX_CHANNEL];
};

/* Global access to xgene DMA device structure */
struct xgene_dma *pdata;

static const char * const xgene_dma_desc_err[] = {
	[ERR_DESC_AXI] = "AXI error when reading src/dst link list",
	[ERR_BAD_DESC] = "ERR or El_ERR fields not set to zero in desc",
	[ERR_READ_DATA_AXI] = "AXI error when reading data",
	[ERR_WRITE_DATA_AXI] = "AXI error when writing data",
	[ERR_FBP_TIMEOUT] = "Timeout on bufpool fetch",
	[ERR_ECC] = "ECC double bit error",
	[ERR_DIFF_SIZE] = "Bufpool too small to hold all the DIF result",
	[ERR_SCT_GAT_LEN] = "Gather and scatter data length not same",
	[ERR_CRC_ERR] = "CRC error",
	[ERR_CHKSUM] = "Checksum error",
	[ERR_DIF] = "DIF error",
};

static const char * const xgene_dma_err[] = {
	[ERR_DIF_SIZE_INT] = "DIF size error",
	[ERR_GS_ERR_INT] = "Gather scatter not same size error",
	[ERR_FPB_TIMEO_INT] = "Free pool time out error",
	[ERR_WFIFO_OVF_INT] = "Write FIFO over flow error",
	[ERR_RFIFO_OVF_INT] = "Read FIFO over flow error",
	[ERR_WR_TIMEO_INT] = "Write time out error",
	[ERR_RD_TIMEO_INT] = "Read time out error",
	[ERR_WR_ERR_INT] = "HBF bus write error",
	[ERR_RD_ERR_INT] = "HBF bus read error",
	[ERR_BAD_DESC_INT] = "Ring descriptor HE0 not set error",
	[ERR_DESC_DST_INT] = "HFB reading dst link address error",
	[ERR_DESC_SRC_INT] = "HFB reading src link address error",
};

static void xgene_dma_desc_dump(void *desc, const char *str)
{
	print_hex_dump(KERN_ERR, str, DUMP_PREFIX_ADDRESS,
		       16, 8, desc, 32, true);
	pr_info("\n");
}

void xgene_dma_queue_dump(struct xgene_qmtm_qinfo *ring)
{
	char msg[16];
	int i;

	pr_info("=================================================================\n");
	pr_info("Queue id %d PBN %d vaddr 0x%p paddr 0x%llX\n"
	       "slot count %d command 0x%p IRQ %d Current head %d\n",
	       ring->queue_id, ring->pbn, ring->queue_vaddr, ring->queue_paddr,
	       ring->count, ring->command, ring->irq, ring->qhead);
	pr_info("=================================================================\n\n");

	for (i = 0; i < ring->count; i++) {
		sprintf(msg, "QMSG%d: ", i);
		xgene_dma_desc_dump(&ring->msg32[i], msg);
	}
}

/**
 * xgene_dma_chan_hw_queue_desc_dump - Function to dump Tx and Rx
 * hw queue descriptors of specific DMA channel.
 * @chan_id: DMA channel id (0 - 3)
 */
void xgene_dma_chan_hw_queue_desc_dump(int chan_id)
{
	struct xgene_dma_chan *chan = &pdata->chan[chan_id];

	pr_info("Channel %d: TX queue descriptor Dump:\n", chan_id);
	xgene_dma_queue_dump(&chan->tx_ring);

	pr_info("Channel %d: Rx queue descriptor Dump:\n", chan_id);
	xgene_dma_queue_dump(&chan->rx_ring);
}
EXPORT_SYMBOL_GPL(xgene_dma_chan_hw_queue_desc_dump);

/*
 * Function to dump Tx and Rx hw queue descriptors for all
 * DMA channels.
 */
void xgene_dma_hw_queue_desc_dump_all(void)
{
	int i;

	pr_info("PktDMA: HW queue descriptor Dump\n");
	pr_info("================================\n");

	for (i = 0; i < DMA_MAX_CHANNEL; i++)
		xgene_dma_chan_hw_queue_desc_dump(i);

	pr_info("PktDMA: End of HW queue descriptor Dump\n");
	pr_info("=======================================\n");
}
EXPORT_SYMBOL_GPL(xgene_dma_hw_queue_desc_dump_all);

static void xgene_dma_ld_dump(struct list_head *head)
{
	struct xgene_dma_desc_sw *desc_sw;
	int j = 0;

	list_for_each_entry(desc_sw, head, node) {
		xgene_dma_desc_dump(&desc_sw->desc1, "TX QMSG1: ");

		if (desc_sw->flags & DMA_FLAG_64B_DESC)
			xgene_dma_desc_dump(&desc_sw->desc2, "TX QMSG2: ");
		j++;
	}

	if (j)
		pr_info("LD count %d\n", j);
	else
		pr_info("No LDs\n");
}

/**
 * xgene_dma_chan_sw_ld_queue_dump - Function to dump prepared, submitted,
 * and running descriptors. This will dump ld pending, ld running,
 * ld completed etc.
 * @chan_id: DMA channel id (0 - 3)
 */
void xgene_dma_chan_sw_ld_queue_dump(int chan_id)
{
	struct xgene_dma_chan *chan = &pdata->chan[chan_id];

	/*
	 * Dump the descriptor which has been prepared and submitted
	 * by client, but not yet queued to DMA hw queue descriptor
	 * for execution.
	 */
	pr_info("Channel %d Pending link descriptor Dump:\n", chan_id);
	xgene_dma_ld_dump(&chan->ld_pending);

	/* Dump the descriptor which has been prepared, submitted and queued
	 * to DMA hw queue descriptor for execution, but not yet completed by
	 * DMA engine and still waiting for comletion.
	 */
	pr_info("Channel %d Running link descriptor Dump:\n", chan_id);
	xgene_dma_ld_dump(&chan->ld_running);

	/* Dump the descriptor which has been executed and processed by driver,
	 * but not yet free, still waiting for client to DMA_CTRL_ACK this tx
	 * descriptor to release it from DMA pool. Client has to acknowledge
	 * this to free by DMA driver if he is no more using this tx descriptor.
	 */
	pr_info("Channel %d Completed link descriptor Dump:\n", chan_id);
	xgene_dma_ld_dump(&chan->ld_completed);
}
EXPORT_SYMBOL_GPL(xgene_dma_chan_sw_ld_queue_dump);

/*
 * Function to dump prepared, submitted, and running descriptors.
 * This will dump for all DMA channel's sw queue link list e.g. ld pending,
 * ld running, ld completed etc.
 */
void xgene_dma_sw_ld_queue_dump_all(void)
{
	int i;

	pr_info("PktDMA: SW link descriptor Dump\n");
	pr_info("===============================\n");

	for (i = 0; i < DMA_MAX_CHANNEL; i++)
		xgene_dma_chan_sw_ld_queue_dump(i);

	pr_info("PktDMA: End of SW link descriptor Dump\n");
	pr_info("======================================\n");
}
EXPORT_SYMBOL_GPL(xgene_dma_sw_ld_queue_dump_all);

/* Check for PQ operation enable/disable (on Channel 1) in hw */
static bool is_pq_enabled(struct xgene_dma *pdma)
{
	u32 val;

	val = ioread32(pdma->csr_efuse + SOC_JTAG1_SHADOW);
	return !(val & PQ_DISABLE_MASK);
}

static bool xgene_dma_is_xor_enabled(struct xgene_dma_chan *chan)
{
#ifdef DMA_ENABLE_XOR
	if (((chan->id == DMA_XOR_CHANNEL) && !is_pq_enabled(chan->pdma)) ||
	    ((chan->id == DMA_PQ_CHANNEL) && is_pq_enabled(chan->pdma)))
		return 1;

	return 0;
#else
	return 0;
#endif
}

static bool xgene_dma_is_pq_enabled(struct xgene_dma_chan *chan)
{
#ifdef DMA_ENABLE_PQ
	if ((chan->id == DMA_PQ_CHANNEL) && is_pq_enabled(chan->pdma))
		return 1;
	return 0;
#else
	return 0;
#endif
}

static bool xgene_dma_is_memcpy_enabled(struct xgene_dma_chan *chan)
{
#ifdef DMA_ENABLE_MEMCPY_ALL_CHANNEL
	return 1;
#elif defined(DMA_ENABLE_MEMCPY)
	if (chan->id == DMA_COPY_CHANNEL)
		return 1;
	return 0;
#else
	return 0;
#endif
}

static bool xgene_dma_is_sg_enabled(struct xgene_dma_chan *chan)
{
#ifdef DMA_ENABLE_SG_ALL_CHANNEL
	return 1;
#elif defined(DMA_ENABLE_SG)
	if (chan->id == DMA_COPY_CHANNEL)
		return 1;
	return 0;
#else
	return 0;
#endif
}

static void xgene_dma_cpu_to_le64(u64 *desc, int count)
{
	int i;

	for (i = 0; i < count; i++)
		desc[i] = cpu_to_le64(desc[i]);
}

static void xgene_dma_le64_to_cpu(u64 *desc, int count)
{
	int i;

	for (i = 0; i < count; i++)
		desc[i] = le64_to_cpu(desc[i]);
}

static u16 xgene_dma_encode_len(u32 len)
{
	return (len < DMA_MAX_BYTE_CNT) ?
		len : DMA_16K_BUFFER_LEN_CODE;
}

static struct xgene_dma_chan *xgene_dma_flyby_channel(void)
{
	return &pdata->chan[DMA_FLYBY_CHANNEL];
}

static u8 xgene_dma_encode_xor_flyby(u32 src_cnt)
{
	static u8 flyby_type[] = {
		FLYBY_2SRC_XOR, /* Dummy */
		FLYBY_2SRC_XOR, /* Dummy */
		FLYBY_2SRC_XOR,
		FLYBY_3SRC_XOR,
		FLYBY_4SRC_XOR,
		FLYBY_5SRC_XOR
	};

	return flyby_type[src_cnt];
}

static u32 xgene_dma_ring_desc_cnt(struct xgene_qmtm_qinfo *ring)
{
	u32 __iomem *qfabric = ring->qfabric;
	u32 ring_state = ioread32(&qfabric[1]);

	return (ring_state & DMA_RING_DESC_CNT_MASK) >>
		DMA_RING_DESC_CNT_POS;
}

static void xgene_dma_set_src_buffer(void *ext8, size_t *len,
				     dma_addr_t *paddr)
{
	size_t nbytes = (*len < DMA_MAX_BYTE_CNT) ?
			*len : DMA_MAX_BYTE_CNT;

	DMA_DESC_BUFADDR_SET(ext8, *paddr);
	DMA_DESC_BUFLEN_SET(ext8, xgene_dma_encode_len(nbytes));
	*len -= nbytes;
	*paddr += nbytes;
}

static void xgene_dma_invalidate_buffer(void *ext8)
{
	DMA_DESC_BUFLEN_SET(ext8, DMA_INVALID_LEN_CODE);
}

static void *xgene_dma_lookup_ext8(u64 *desc, int idx)
{
	return (idx % 2) ? (desc + idx - 1) : (desc + idx + 1);
}

static void xgene_dma_init_desc(void *desc, u32 queue_id)
{
	DMA_DESC_C_SET(desc); /* Coherent IO */
	DMA_DESC_IN_SET(desc);
	DMA_DESC_H0ENQ_NUM_SET(desc, QMTM_QUEUE_ID(DMA_QMTM_IP, queue_id));
	DMA_DESC_RTYPE_SET(desc, DMA_RING_OWNER);
}

static void xgene_dma_prep_cpy_desc(struct xgene_dma_chan *chan,
				    struct xgene_dma_desc_sw *desc_sw,
				    dma_addr_t dst, dma_addr_t src,
				    size_t len)
{
	void *desc1, *desc2;
	int i;

	/* Get 1st descriptor */
	desc1 = &desc_sw->desc1;
	xgene_dma_init_desc(desc1, chan->rx_ring.queue_id);

	/* Set destination address */
	DMA_DESC_DR_SET(desc1);
	DMA_DESC_DST_ADDR_SET(desc1, dst);

	/* Set 1st source address */
	xgene_dma_set_src_buffer(desc1 + 8, &len, &src);

	if (len <= 0) {
		desc2 = NULL;
		goto skip_additional_src;
	}

	/*
	 * We need to split this source buffer,
	 * and need to use 2nd descriptor
	 */
	desc2 = &desc_sw->desc2;
	DMA_DESC_NV_SET(desc1);

	/* Set 2nd to 5th source address */
	for (i = 0; i < 4 && len; i++)
		xgene_dma_set_src_buffer(xgene_dma_lookup_ext8(desc2, i),
					 &len, &src);

	/* Invalidate unused source address field */
	for (; i < 4; i++)
		xgene_dma_invalidate_buffer(xgene_dma_lookup_ext8(desc2, i));

	/* Update flag that we have prepared 64B descriptor */
	desc_sw->flags |= DMA_FLAG_64B_DESC;

skip_additional_src:
	/* Hardware stores descriptor in little endian format */
	xgene_dma_cpu_to_le64(desc1, 4);
	if (desc2)
		xgene_dma_cpu_to_le64(desc2, 4);
}

static void xgene_dma_prep_xor_desc(struct xgene_dma_chan *chan,
				    struct xgene_dma_desc_sw *desc_sw,
				    dma_addr_t *dst, dma_addr_t *src,
				    u32 src_cnt, size_t *nbytes,
				    const u8 *scf)
{
	void *desc1, *desc2;
	size_t len = *nbytes;
	int i;

	desc1 = &desc_sw->desc1;
	desc2 = &desc_sw->desc2;

	/* Initialize DMA descriptor */
	xgene_dma_init_desc(desc1, chan->rx_ring.queue_id);

	/* Set destination address */
	DMA_DESC_DR_SET(desc1);
	DMA_DESC_DST_ADDR_SET(desc1, *dst);

	/* We have multiple source addresses, so need to set NV bit*/
	DMA_DESC_NV_SET(desc1);

	/* Set flyby opcode */
	DMA_DESC_FLYBY_SET(desc1, xgene_dma_encode_xor_flyby(src_cnt));

	/* Set 1st to 5th source addresses */
	for (i = 0; i < src_cnt; i++) {
		len = *nbytes;
		xgene_dma_set_src_buffer((i == 0) ? (desc1 + 8) :
					 xgene_dma_lookup_ext8(desc2, i - 1),
					 &len, &src[i]);
		DMA_DESC_MULTI_SET(desc1, scf[i], i);
	}

	/* Hardware stores descriptor in little endian format */
	xgene_dma_cpu_to_le64(desc1, 4);
	xgene_dma_cpu_to_le64(desc2, 4);

	/* Update meta data */
	*nbytes = len;
	*dst += DMA_MAX_BYTE_CNT;

	/* We need always 64B descriptor to perform xor or pq operations */
	desc_sw->flags |= DMA_FLAG_64B_DESC;
}

static u32 xgene_dma_set_flyby_src(void *ext8, struct scatterlist *sg,
				   dma_addr_t *paddr, u32 *nbytes, u32 offset)
{
	u32 len;

	/* Fetch physical address from sg */
	if (*paddr == 0)
		*paddr = sg_dma_address(sg);

	len = sg_dma_len(sg) - offset;

	DMA_DESC_BUFADDR_SET(ext8, *paddr);
	DMA_DESC_BUFLEN_SET(ext8, xgene_dma_encode_len(len));

	if (len <= DMA_MAX_BYTE_CNT) {
		*nbytes -= len;
		*paddr = 0;
		return len;
	}

	*nbytes -= DMA_MAX_BYTE_CNT;
	*paddr += DMA_MAX_BYTE_CNT;
	return DMA_MAX_BYTE_CNT;
}

static int xgene_dma_prep_flyby_desc(struct xgene_dma_chan *chan,
				     struct xgene_dma_desc_sw *desc_sw,
				     struct scatterlist *sg, u32 nbytes,
				     u32 seed, u8 opcode)
{
	void *desc1, *desc2;
	dma_addr_t paddr = 0;
	u32 len = nbytes;
	u32 offset = 0;
	int i;

	/* Get 1st descriptor */
	desc1 = &desc_sw->desc1;
	xgene_dma_init_desc(desc1, chan->rx_ring.queue_id);

	/* Set 1st source address */
	offset += xgene_dma_set_flyby_src(desc1 + 8, sg, &paddr,
					  &nbytes, offset);

	if (!nbytes) {
		desc2 = NULL;
		goto skip_additional_src;
	}

	/*
	 * Still we have request length remaining,
	 * So we need to use prepare 64B descriptor
	 */
	desc2 = &desc_sw->desc2;
	DMA_DESC_NV_SET(desc1);

	/* Set 2nd to 5th source address */
	for (i = 0; i < 4 && nbytes; i++) {
		/* Fetch next sg element */
		if (!paddr) {
			sg = sg_next(sg);
			if (!sg)
				break;
			offset = 0;
		}
		offset += xgene_dma_set_flyby_src(
				xgene_dma_lookup_ext8(desc2, i),
				sg, &paddr, &nbytes, offset);
	}

	/* Invalidate unused source address field */
	for (; i < 4; i++)
		xgene_dma_invalidate_buffer(xgene_dma_lookup_ext8(desc2, i));

	/* Check whether requested buffer processed */
	if (nbytes) {
		chan_err(chan, "Src count crossed maximum limit\n");
		return -EINVAL;
	}

	/* Update flag that we have prepared 64B descriptor */
	desc_sw->flags |= DMA_FLAG_64B_DESC;

skip_additional_src:
	/* Set descriptor parameters for flyby operation */
	DMA_DESC_BD_SET(desc1);
	DMA_DESC_SD_SET(desc1);
	DMA_DESC_FLYBY_SET(desc1, opcode);
	DMA_DESC_CRCBYTECNT_SET(desc1, len);
	DMA_DESC_CRCSEED_SET(desc1, bitrev32(seed));

	/* Hardware stores descriptor in little endian format */
	xgene_dma_cpu_to_le64(desc1, 4);
	if (desc2)
		xgene_dma_cpu_to_le64(desc2, 4);

	return 0;
}

static dma_cookie_t xgene_dma_tx_submit(struct dma_async_tx_descriptor *tx)
{
	struct xgene_dma_desc_sw *desc;
	struct xgene_dma_chan *chan;
	dma_cookie_t cookie;

	BUG_ON(!tx);

	chan = to_dma_chan(tx->chan);
	desc = to_dma_desc_sw(tx);

	spin_lock_bh(&chan->lock);

	cookie = dma_cookie_assign(tx);

	/* Add this transaction list onto the tail of the pending queue */
	list_splice_tail_init(&desc->tx_list, &chan->ld_pending);

	spin_unlock_bh(&chan->lock);

	return cookie;
}

static void xgene_dma_free_descriptor(struct xgene_dma_chan *chan,
				      struct xgene_dma_desc_sw *desc)
{
	list_del(&desc->node);
	chan_dbg(chan, "LD %p free\n", desc);
	dma_pool_free(chan->desc_pool, desc, desc->tx.phys);
}

static struct xgene_dma_desc_sw *xgene_dma_alloc_descriptor(
				 struct xgene_dma_chan *chan)
{
	struct xgene_dma_desc_sw *desc;
	dma_addr_t phys;

	desc = dma_pool_alloc(chan->desc_pool, GFP_NOWAIT, &phys);
	if (!desc) {
		chan_err(chan, "Failed to allocate LDs\n");
		return NULL;
	}

	memset(desc, 0, sizeof(*desc));

	INIT_LIST_HEAD(&desc->tx_list);
	desc->tx.phys = phys;
	desc->tx.tx_submit = xgene_dma_tx_submit;
	dma_async_tx_descriptor_init(&desc->tx, &chan->dma_chan);

	chan_dbg(chan, "LD %p allocated\n", desc);

	return desc;
}

/**
 * xgene_dma_clean_completed_descriptor - free all descriptors which
 * has been completed and acked
 * @chan: DMA channel
 *
 * This function is used on all completed and acked descriptors.
 * All descriptors should only be freed in this function.
 */
static void xgene_dma_clean_completed_descriptor(struct xgene_dma_chan *chan)
{
	struct xgene_dma_desc_sw *desc, *_desc;

	/* Run the callback for each descriptor, in order */
	list_for_each_entry_safe(desc, _desc, &chan->ld_completed, node) {
		if (async_tx_test_ack(&desc->tx))
			xgene_dma_free_descriptor(chan, desc);
	}
}

static void xgene_dma_unmap_dst(struct xgene_dma_chan *chan,
				struct xgene_dma_desc_sw *desc)
{
	enum dma_data_direction dir;

	if (desc->src_cnt > 1) /* is xor, pq? */
		dir = DMA_BIDIRECTIONAL;
	else
		dir = DMA_FROM_DEVICE;

	if (desc->tx.flags & DMA_COMPL_DEST_UNMAP_SINGLE) {
		if (desc->flags & DMA_FLAG_FIRST_DST)
			dma_unmap_single(chan->dev,
					 desc->dst[0], desc->len, dir);
		if (desc->flags & DMA_FLAG_SECOND_DST)
			dma_unmap_single(chan->dev,
					 desc->dst[1], desc->len, dir);
	} else {
		if (desc->flags & DMA_FLAG_FIRST_DST)
			dma_unmap_page(chan->dev,
				       desc->dst[0], desc->len, dir);
		if (desc->flags & DMA_FLAG_SECOND_DST)
			dma_unmap_page(chan->dev,
				       desc->dst[1], desc->len, dir);
	}
}

static void xgene_dma_unmap_src(struct xgene_dma_chan *chan,
				struct xgene_dma_desc_sw *desc)
{
	int i;

	for (i = 0; i < desc->src_cnt; i++) {
		if (desc->dst[0] == desc->src[0])
			continue;

		if (desc->tx.flags & DMA_COMPL_SRC_UNMAP_SINGLE)
			dma_unmap_single(chan->dev, desc->src[i],
					 desc->len, DMA_TO_DEVICE);
		else
			dma_unmap_page(chan->dev, desc->src[i],
				       desc->len, DMA_TO_DEVICE);
	}
}

static void xgene_dma_tx_descriptor_unmap(struct xgene_dma_chan *chan,
					  struct xgene_dma_desc_sw *desc)
{
	if (desc->flags & DMA_FLAG_SG_ACTIVE)
		return;

	if (!(desc->tx.flags & DMA_COMPL_SKIP_DEST_UNMAP))
		xgene_dma_unmap_dst(chan, desc);

	if (!(desc->tx.flags & DMA_COMPL_SKIP_SRC_UNMAP))
		xgene_dma_unmap_src(chan, desc);
}

/**
 * xgene_dma_run_tx_complete_actions - cleanup a single link descriptor
 * @chan: DMA channel
 * @desc: descriptor to cleanup and free
 *
 * This function is used on a descriptor which has been executed by the DMA
 * controller. It will run any callbacks, submit any dependencies.
 */
static void xgene_dma_run_tx_complete_actions(struct xgene_dma_chan *chan,
					      struct xgene_dma_desc_sw *desc_sw,
					      void *desc_hw)
{
	struct dma_async_tx_descriptor *tx = &desc_sw->tx;

	/* Run the link descriptor callback function */
	if (desc_sw->flags & DMA_FLAG_FLYBY_ACTIVE) {
		if (desc_sw->flyby_cb)
			desc_sw->flyby_cb(desc_sw->flyby_ctx,
					  DMA_DESC_CRC_RESULT_RD(desc_hw));
		/* Unmap flyb source SG */
		dma_unmap_sg(chan->dev,
			     desc_sw->src_sg, desc_sw->src_cnt, DMA_TO_DEVICE);
		return;
	}

	/*
	 * If this is not the last transaction in the group,
	 * then no need to complete cookie and run any callback as
	 * this is not the tx_descriptor which had been sent to caller
	 * of this DMA request
	 */
	if (tx->cookie == 0)
		return;

	dma_cookie_complete(tx);

	if (tx->callback)
		tx->callback(tx->callback_param);

	/* Unmap destinations and sources if requested by client */
	xgene_dma_tx_descriptor_unmap(chan, desc_sw);

	/* Run any dependencies */
	dma_run_dependencies(tx);
}

/**
 * xgene_dma_clean_running_descriptor - move the completed descriptor from
 * ld_running to ld_completed
 * @chan: DMA channel
 * @desc: the descriptor which is completed
 *
 * Free the descriptor directly if acked by async_tx api and this driver,
 * or move it to queue ld_completed.
 */
static void xgene_dma_clean_running_descriptor(struct xgene_dma_chan *chan,
					       struct xgene_dma_desc_sw *desc)
{
	/* Remove from the list of transactions */
	list_del(&desc->node);

	/*
	 * the client is allowed to attach dependent operations
	 * until 'ack' is set
	 */
	if (async_tx_test_ack(&desc->tx) ||
	    (desc->flags & DMA_FLAG_FLYBY_ACTIVE)) {
		chan_dbg(chan, "LD %p free\n", desc);
		dma_pool_free(chan->desc_pool, desc, desc->tx.phys);
	} else {
		/*
		 * Move this descriptor to the list of descriptors which is
		 * completed, but still awaiting the 'ack' bit to be set.
		 */
		list_add_tail(&desc->node, &chan->ld_completed);
	}
}

static int xgene_chan_xfer_request(struct xgene_dma_chan *chan,
				   struct xgene_dma_desc_sw *desc_sw)
{
	struct xgene_qmtm_qinfo *ring = &chan->tx_ring;
	struct xgene_qmtm_msg32 *desc_hw;

	/* Check if can push more descriptor to hw for execution */
	if (xgene_dma_ring_desc_cnt(ring) > (ring->count - 2))
		return -EBUSY;

	/* Get hw descriptor from DMA tx ring */
	desc_hw = &ring->msg32[ring->qhead];

	/*
	 * Increment the head count to point next
	 * descriptor for next time
	 */
	if (++ring->qhead == ring->count)
		ring->qhead = 0;

	/* Copy prepared sw descriptor data to hw descriptor */
	memcpy(desc_hw, &desc_sw->desc1, sizeof(*desc_hw));

	/*
	 * Check if we have prepared 64B descriptor,
	 * in this case we need one more hw descriptor
	 */
	if (desc_sw->flags & DMA_FLAG_64B_DESC) {
		desc_hw = &ring->msg32[ring->qhead];

		if (++ring->qhead == ring->count)
			ring->qhead = 0;

		memcpy(desc_hw, &desc_sw->desc2, sizeof(*desc_hw));
	}

	/* Notify the hw that we have descriptor ready for execution */
	iowrite32((desc_sw->flags & DMA_FLAG_64B_DESC) ? 2 : 1, ring->command);

	return 0;
}

/**
 * xgene_chan_xfer_ld_pending - push any pending transactions to hw
 * @chan : DMA channel
 *
 * LOCKING: must hold chan->lock
 */
static void xgene_chan_xfer_ld_pending(struct xgene_dma_chan *chan)
{
	struct xgene_dma_desc_sw *desc_sw, *_desc_sw;
	int ret;

	/*
	 * If the list of pending descriptors is empty, then we
	 * don't need to do any work at all
	 */
	if (list_empty(&chan->ld_pending)) {
		chan_dbg(chan, "No pending LDs\n");
		return;
	}

	/*
	 * Move elements from the queue of pending transactions onto the list
	 * of running transactions and push it to hw for further executions
	 */
	list_for_each_entry_safe(desc_sw, _desc_sw, &chan->ld_pending, node) {
		/*
		 * Check if have pushed max number of transactions to hw
		 * as capable, so let's stop here and will push remaining
		 * elements from pening ld queue after completing some
		 * descriptors that we have already pushed
		 */
		if (chan->pending >= chan->max_outstanding)
			return;

		ret = xgene_chan_xfer_request(chan, desc_sw);
		if (ret)
			return;

		/*
		 * Delete this element from ld pending queue and append it to
		 * ld running queue
		 */
		list_move_tail(&desc_sw->node, &chan->ld_running);

		/* Increment the pending transaction count */
		chan->pending++;
	}
}

/**
 * xgene_dma_cleanup_descriptors - cleanup link descriptors which are completed
 * and move them to ld_completed to free until flag 'ack' is set
 * @chan: DMA channel
 *
 * This function is used on descriptors which have been executed by the DMA
 * controller. It will run any callbacks, submit any dependencies, then
 * free these descriptors if flag 'ack' is set. Client must not submit or
 * prepare any new requests.
 */
static void xgene_dma_cleanup_descriptors(struct xgene_dma_chan *chan)
{
	struct xgene_qmtm_qinfo *ring = &chan->rx_ring;
	struct xgene_dma_desc_sw *desc_sw, *_desc_sw;
	void *desc_hw1, *desc_hw2;
	u32 command;
	u8 status;

	/* Clean already completed and acked descriptors */
	xgene_dma_clean_completed_descriptor(chan);

	/* Run the callback for each descriptor, in order */
	list_for_each_entry_safe(desc_sw, _desc_sw, &chan->ld_running, node) {
		/* Get subsequent hw descriptor from DMA rx ring */
		desc_hw1 = &ring->msg32[ring->qhead];

		/* Check if this descriptor has been completed */
		if (unlikely(DMA_DESC_IS_EMPTY(desc_hw1)))
			break;

		xgene_dma_le64_to_cpu(desc_hw1, 4);

		if (++ring->qhead == ring->count)
			ring->qhead = 0;

		if (DMA_DESC_NV_RD(desc_hw1)) {
			/* 64B Rx descriptor */
			desc_hw2 = &ring->msg32[ring->qhead];

			if (++ring->qhead == ring->count)
				ring->qhead = 0;

			command = 2;
		} else {
			desc_hw2 = NULL;
			command = 1;
		}

		/* Check if we have any error with DMA transactions */
		status = DMA_DESC_STATUS(DMA_DESC_ELERR_RD(desc_hw1),
					 DMA_DESC_LERR_RD(desc_hw1));
		if (status) {
			/* Print the DMA error type */
			chan_err(chan, "%s\n", xgene_dma_desc_err[status]);

			/*
			 * We have DMA transactions error here. Dump DMA Tx
			 * and Rx descriptors for this request */
			xgene_dma_desc_dump(&desc_sw->desc1, "TX QMSG1:");

			if (desc_sw->flags & DMA_FLAG_64B_DESC)
				xgene_dma_desc_dump(&desc_sw->desc2,
						    "TX QMSG2: ");

			xgene_dma_desc_dump(desc_hw1, "RX ERR QMSG1: ");
			if (desc_hw2)
				xgene_dma_desc_dump(desc_hw2, "RX ERR QMSG2: ");
		}

		/* Notify the hw about this completed descriptor */
		if (desc_sw->flags & DMA_FLAG_FLYBY_ACTIVE)
			iowrite32(command, chan->bufpool.command);

		iowrite32(-command, ring->command);

		xgene_dma_run_tx_complete_actions(chan, desc_sw, desc_hw1);

		xgene_dma_clean_running_descriptor(chan, desc_sw);

		/* Mark this hw descriptor as processed */
		DMA_DESC_SET_EMPTY(desc_hw1);
		if (desc_hw2)
			DMA_DESC_SET_EMPTY(desc_hw2);

		/*
		 * Decrement the pending transaction count
		 * as we have processed one
		 */
		chan->pending--;
	}

	/*
	 * Start any pending transactions automatically
	 *
	 * In the ideal case, we keep the DMA controller busy while we go
	 * ahead and free the descriptors below.
	 */
	xgene_chan_xfer_ld_pending(chan);
}

static int xgene_dma_alloc_chan_resources(struct dma_chan *dchan)
{
	struct xgene_dma_chan *chan = to_dma_chan(dchan);

	/* Has this channel already been allocated? */
	if (chan->desc_pool)
		return 1;

	chan->desc_pool = dma_pool_create(chan->name, chan->dev,
					  sizeof(struct xgene_dma_desc_sw),
					  0, 0);
	if (!chan->desc_pool) {
		chan_err(chan, "Failed to allocate descriptor pool\n");
		return -ENOMEM;
	}

	chan_dbg(chan, "Allocate descriptor pool\n");

	return 1;
}

/**
 * xgene_dma_free_desc_list - Free all descriptors in a queue
 * @chan: DMA channel
 * @list: the list to free
 *
 * LOCKING: must hold chan->lock
 */
static void xgene_dma_free_desc_list(struct xgene_dma_chan *chan,
				     struct list_head *list)
{
	struct xgene_dma_desc_sw *desc, *_desc;

	list_for_each_entry_safe(desc, _desc, list, node)
		xgene_dma_free_descriptor(chan, desc);
}

static void xgene_dma_free_desc_list_reverse(struct xgene_dma_chan *chan,
					     struct list_head *list)
{
	struct xgene_dma_desc_sw *desc, *_desc;

	list_for_each_entry_safe_reverse(desc, _desc, list, node)
		xgene_dma_free_descriptor(chan, desc);
}

static void xgene_dma_free_chan_resources(struct dma_chan *dchan)
{
	struct xgene_dma_chan *chan = to_dma_chan(dchan);

	chan_dbg(chan, "Free all resources\n");

	if (!chan->desc_pool)
		return;

	spin_lock_bh(&chan->lock);

	/* Process all running descriptor */
	xgene_dma_cleanup_descriptors(chan);

	/* Clean all link descriptor queues */
	xgene_dma_free_desc_list(chan, &chan->ld_pending);
	xgene_dma_free_desc_list(chan, &chan->ld_running);
	xgene_dma_free_desc_list(chan, &chan->ld_completed);

	spin_unlock_bh(&chan->lock);

	/* Delete this channel DMA pool */
	dma_pool_destroy(chan->desc_pool);
	chan->desc_pool = NULL;
}

int xgene_dma_flyby_alloc_resources(void)
{
	struct xgene_dma_chan *chan = xgene_dma_flyby_channel();
	struct dma_chan *dchan = &chan->dma_chan;

	spin_lock_bh(&chan->lock);

	/* Increment client count */
	dchan->client_count++;
	spin_unlock_bh(&chan->lock);

	/* allocate upon first client reference */
	if (dchan->client_count == 1) {
		int desc_cnt = xgene_dma_alloc_chan_resources(dchan);

		if (desc_cnt < 0) {
			dchan->client_count = 0;
			return -ENOMEM;
		}
	}

	return 0;
}
EXPORT_SYMBOL_GPL(xgene_dma_flyby_alloc_resources);

void xgene_dma_flyby_free_resources(void)
{
	struct xgene_dma_chan *chan = xgene_dma_flyby_channel();
	struct dma_chan *dchan = &chan->dma_chan;

	if (!dchan->client_count)
		return; /* this channel failed alloc_chan_resources */

	spin_lock_bh(&chan->lock);

	/* Decrement client count */
	dchan->client_count--;
	spin_unlock_bh(&chan->lock);

	if (dchan->client_count == 0)
		xgene_dma_free_chan_resources(dchan);
}
EXPORT_SYMBOL_GPL(xgene_dma_flyby_free_resources);

static void xgene_dma_issue_pending(struct dma_chan *dchan)
{
	struct xgene_dma_chan *chan = to_dma_chan(dchan);

	spin_lock_bh(&chan->lock);
	xgene_chan_xfer_ld_pending(chan);
	spin_unlock_bh(&chan->lock);
}

static struct dma_async_tx_descriptor *xgene_dma_prep_memcpy(
	struct dma_chan *dchan, dma_addr_t dst, dma_addr_t src,
	size_t len, unsigned long flags)
{
	struct xgene_dma_desc_sw *first = NULL, *new;
	struct xgene_dma_chan *chan;
	dma_addr_t _dst, _src;
	size_t copy, _len;

	BUG_ON(!dchan);
	BUG_ON(!len);

	/* To use in shadowcopy */
	_dst = dst;
	_src = src;
	_len = len;

	chan = to_dma_chan(dchan);

	do {
		/* Allocate the link descriptor from DMA pool */
		new = xgene_dma_alloc_descriptor(chan);
		if (!new)
			goto fail;

		/* Create the largest transaction possible */
		copy = min_t(size_t, len, DMA_MAX_64B_DESC_BYTE_CNT);

		/* Prepare DMA descriptor */
		xgene_dma_prep_cpy_desc(chan, new, dst, src, copy);

		if (!first)
			first = new;

		new->tx.cookie = 0;
		async_tx_ack(&new->tx);

		/* Update metadata */
		len -= copy;
		dst += copy;
		src += copy;

		/* Insert the link descriptor to the LD ring */
		list_add_tail(&new->node, &first->tx_list);
	} while (len);

	/* Shadowcopy to use while unmapping resources */
	new->dst[0] = _dst;
	new->src[0] = _src;
	new->len = _len;
	new->src_cnt = 1;
	new->tx.flags = flags; /* client is in control of this ack */
	new->tx.cookie = -EBUSY;
	new->flags |= DMA_FLAG_FIRST_DST;
	list_splice(&first->tx_list, &new->tx_list);

	return &new->tx;

fail:
	if (!first)
		return NULL;

	xgene_dma_free_desc_list_reverse(chan, &first->tx_list);
	return NULL;
}

static struct dma_async_tx_descriptor *xgene_dma_prep_sg(
	struct dma_chan *dchan, struct scatterlist *dst_sg,
	u32 dst_nents, struct scatterlist *src_sg,
	u32 src_nents, unsigned long flags)
{
	struct xgene_dma_desc_sw *first = NULL, *new = NULL;
	struct xgene_dma_chan *chan;
	size_t dst_avail, src_avail;
	dma_addr_t dst, src;
	size_t len;

	BUG_ON(!dchan);
	BUG_ON(!dst_nents || !src_nents);
	BUG_ON(!dst_sg || !src_sg);

	chan = to_dma_chan(dchan);

	/* Get prepared for the loop */
	dst_avail = sg_dma_len(dst_sg);
	src_avail = sg_dma_len(src_sg);
	dst_nents--;
	src_nents--;

	/* Run until we are out of scatterlist entries */
	while (true) {
		/* Create the largest transaction possible */
		len = min_t(size_t, src_avail, dst_avail);
		len = min_t(size_t, len, DMA_MAX_64B_DESC_BYTE_CNT);
		if (len == 0)
			goto fetch;

		dst = sg_dma_address(dst_sg) + sg_dma_len(dst_sg) - dst_avail;
		src = sg_dma_address(src_sg) + sg_dma_len(src_sg) - src_avail;

		/* Allocate the link descriptor from DMA pool */
		new = xgene_dma_alloc_descriptor(chan);
		if (!new)
			goto fail;

		/* Prepare DMA descriptor */
		xgene_dma_prep_cpy_desc(chan, new, dst, src, len);

		if (!first)
			first = new;

		new->tx.cookie = 0;
		async_tx_ack(&new->tx);

		/* update metadata */
		dst_avail -= len;
		src_avail -= len;

		/* Insert the link descriptor to the LD ring */
		list_add_tail(&new->node, &first->tx_list);

fetch:
		/* fetch the next dst scatterlist entry */
		if (dst_avail == 0) {
			/* no more entries: we're done */
			if (dst_nents == 0)
				break;

			/* fetch the next entry: if there are no more: done */
			dst_sg = sg_next(dst_sg);
			if (!dst_sg)
				break;

			dst_nents--;
			dst_avail = sg_dma_len(dst_sg);
		}

		/* fetch the next src scatterlist entry */
		if (src_avail == 0) {
			/* no more entries: we're done */
			if (src_nents == 0)
				break;

			/* fetch the next entry: if there are no more: done */
			src_sg = sg_next(src_sg);
			if (!src_sg)
				break;

			src_nents--;
			src_avail = sg_dma_len(src_sg);
		}
	}

	BUG_ON(!new);
	new->tx.flags = flags; /* client is in control of this ack */
	new->tx.cookie = -EBUSY;
	new->flags |= DMA_FLAG_SG_ACTIVE;
	list_splice(&first->tx_list, &new->tx_list);

	return &new->tx;

fail:
	if (!first)
		return NULL;

	xgene_dma_free_desc_list_reverse(chan, &first->tx_list);
	return NULL;
}

static struct dma_async_tx_descriptor *xgene_dma_prep_xor(
	struct dma_chan *dchan, dma_addr_t dst, dma_addr_t *src,
	u32 src_cnt, size_t len, unsigned long flags)
{
	struct xgene_dma_desc_sw *first = NULL, *new;
	struct xgene_dma_chan *chan;
	static const u8 multi[DMA_MAX_XOR_SRC] = {
				0x01, 0x01, 0x01, 0x01, 0x01};
	size_t _len;
	dma_addr_t _dst;
	dma_addr_t _src[DMA_MAX_XOR_SRC];

	BUG_ON(!dchan);
	BUG_ON(!len);

	/* To use in shadowcopy */
	_len = len;
	_dst = dst;
	memcpy(_src, src, sizeof(*src) * src_cnt);

	chan = to_dma_chan(dchan);

	do {
		/* Allocate the link descriptor from DMA pool */
		new = xgene_dma_alloc_descriptor(chan);
		if (!new)
			goto fail;

		if (!first)
			first = new;

		/* Prepare xor DMA descriptor */
		xgene_dma_prep_xor_desc(chan, new, &dst, src,
					src_cnt, &len, multi);

		new->tx.cookie = 0;
		async_tx_ack(&new->tx);

		/* Insert the link descriptor to the LD ring */
		list_add_tail(&new->node, &first->tx_list);
	} while (len);

	/* Shadowcopy to use while unmapping resources */
	new->dst[0] = _dst;
	memcpy(new->src, _src, sizeof(*_src) * src_cnt);
	new->len = _len;
	new->src_cnt = src_cnt;
	new->tx.flags = flags; /* client is in control of this ack */
	new->tx.cookie = -EBUSY;
	new->flags |= DMA_FLAG_FIRST_DST;
	list_splice(&first->tx_list, &new->tx_list);

	return &new->tx;

fail:
	if (!first)
		return NULL;

	xgene_dma_free_desc_list_reverse(chan, &first->tx_list);
	return NULL;
}

static struct dma_async_tx_descriptor *xgene_dma_prep_pq(
	struct dma_chan *dchan, dma_addr_t *dst, dma_addr_t *src,
	u32 src_cnt, const u8 *scf, size_t len, unsigned long flags)
{
	struct xgene_dma_desc_sw *first = NULL, *new;
	struct xgene_dma_chan *chan;
	dma_addr_t _src[DMA_MAX_XOR_SRC];
	size_t _len = len;
	u32 desc_flags = 0;
	static const u8 multi[DMA_MAX_XOR_SRC] = {
				0x01, 0x01, 0x01, 0x01, 0x01};
	size_t __len;
	dma_addr_t __dst[2];
	dma_addr_t __src[DMA_MAX_XOR_SRC];

	BUG_ON(!dchan);
	BUG_ON(!len);

	/* To use in shadowcopy */
	__len = len;
	memcpy(__dst, dst, sizeof(*dst) * 2);
	memcpy(__src, src, sizeof(*src) * src_cnt);

	chan = to_dma_chan(dchan);

	/*
	 * Save source addresses on local variable, may be we have to
	 * prepare two descriptor to generate P and Q if both enabled
	 * in the flags by client
	 */
	memcpy(_src, src, sizeof(*src) * src_cnt);

	if (flags & DMA_PREP_PQ_DISABLE_P) {
		len = 0;
		desc_flags |= DMA_FLAG_SECOND_DST;
	}

	if (flags & DMA_PREP_PQ_DISABLE_Q) {
		_len = 0;
		desc_flags |= DMA_FLAG_FIRST_DST;
	}

	do {
		/* Allocate the link descriptor from DMA pool */
		new = xgene_dma_alloc_descriptor(chan);
		if (!new)
			goto fail;

		if (!first)
			first = new;

		new->tx.cookie = 0;
		async_tx_ack(&new->tx);

		/* Insert the link descriptor to the LD ring */
		list_add_tail(&new->node, &first->tx_list);

		/*
		 * Prepare DMA descriptor to generate P,
		 * if DMA_PREP_PQ_DISABLE_P flag is not set
		 */
		if (len) {
			xgene_dma_prep_xor_desc(chan, new, &dst[0], src,
						src_cnt, &len, multi);
			continue;
		}

		/*
		 * Prepare DMA descriptor to generate Q,
		 * if DMA_PREP_PQ_DISABLE_Q flag is not set
		 */
		if (_len) {
			xgene_dma_prep_xor_desc(chan, new, &dst[1], _src,
						src_cnt, &_len, scf);
		}
	} while (len || _len);

	/* Shadowcopy to use in unmap */
	new->len = __len;
	memcpy(new->dst, __dst, sizeof(*__dst) * 2);
	memcpy(new->src, __src, sizeof(*__src) * src_cnt);
	new->src_cnt = src_cnt;
	new->tx.flags = flags; /* client is in control of this ack */
	new->tx.cookie = -EBUSY;
	new->flags |= desc_flags;
	list_splice(&first->tx_list, &new->tx_list);

	return &new->tx;

fail:
	if (!first)
		return NULL;

	xgene_dma_free_desc_list_reverse(chan, &first->tx_list);
	return NULL;
}

int xgene_dma_prep_flyby(dma_flyby_cb flyby_cb, void *ctx,
			 struct scatterlist *src_sg,
			 u32 nbytes, u32 seed, u8 opcode)
{
	struct xgene_dma_chan *chan = xgene_dma_flyby_channel();
	struct xgene_dma_desc_sw *desc;
	int src_cnt, sg_count;
	int ret;

	if (nbytes > DMA_MAX_FLYBY_BYTE_CNT) {
		chan_err(chan,
			 "Flyby source length is too long 0x%X\n", nbytes);
		return -EINVAL;
	}

	src_cnt = sg_nents(src_sg);

	sg_count = dma_map_sg(chan->dev, src_sg, src_cnt, DMA_TO_DEVICE);
	if (!sg_count) {
		chan_err(chan, "Failed to map source sg\n");
		return -ENOMEM;
	}

	/* Allocate the link descriptor from DMA pool */
	desc = xgene_dma_alloc_descriptor(chan);
	if (!desc) {
		chan_err(chan, "Failed to allocate LDs");
		return -ENOMEM;
	}

	/* Prepare DMA flyby descriptor */
	ret = xgene_dma_prep_flyby_desc(chan, desc, src_sg, nbytes,
					seed, opcode);
	if (ret) {
		xgene_dma_free_descriptor(chan, desc);
		return ret;
	}

	/* Setup parameters to use in completion callback */
	desc->flyby_cb = flyby_cb;
	desc->flyby_ctx = ctx;
	desc->src_sg = src_sg;
	desc->src_cnt = src_cnt;
	desc->flags |= DMA_FLAG_FLYBY_ACTIVE;

	spin_lock_bh(&chan->lock);

	/* Add this transaction list onto the tail of the pending queue */
	list_add_tail(&desc->node, &chan->ld_pending);

	/* Start the execution of this transaction */
	xgene_chan_xfer_ld_pending(chan);

	spin_unlock_bh(&chan->lock);

	return -EINPROGRESS;
}
EXPORT_SYMBOL_GPL(xgene_dma_prep_flyby);

static enum dma_status xgene_dma_tx_status(struct dma_chan *dchan,
					   dma_cookie_t cookie,
					   struct dma_tx_state *txstate)
{
	return dma_cookie_status(dchan, cookie, txstate);
}

static void xgene_dma_tasklet_cb(unsigned long data)
{
	struct xgene_dma_chan *chan = (struct xgene_dma_chan *)data;

	spin_lock_bh(&chan->lock);

	/* Run all cleanup for descriptors which have been completed */
	xgene_dma_cleanup_descriptors(chan);

	/* Re-enable DMA channel IRQ */
	enable_irq(chan->rx_irq);

	spin_unlock_bh(&chan->lock);
}

static irqreturn_t xgene_dma_chan_ring_isr(int irq, void *id)
{
	struct xgene_dma_chan *chan = (struct xgene_dma_chan *)id;

	/*
	 * Disable DMA channel IRQ until we process completed
	 * descriptors
	 */
	disable_irq_nosync(chan->rx_irq);

	/*
	 * Schedule the tasklet to handle all cleanup of the current
	 * transaction. It will start a new transaction if there is
	 * one pending.
	 */
	tasklet_schedule(&chan->tasklet);

	return IRQ_HANDLED;
}

static irqreturn_t xgene_dma_err_isr(int irq, void *id)
{
	struct xgene_dma *pdma = (struct xgene_dma *)id;
	unsigned long int_mask;
	u32 val, i;

	val = ioread32(pdma->csr_dma + DMA_INT);

	/* Clear DMA interrupts */
	iowrite32(val, pdma->csr_dma + DMA_INT);

	/* Print DMA error info */
	int_mask = val >> DMA_INT_MASK_SHIFT;
	for_each_set_bit(i, &int_mask, ARRAY_SIZE(xgene_dma_err))
		dev_err(pdma->dev,
			"Interrupt status 0x%08X %s\n", val, xgene_dma_err[i]);

	return IRQ_HANDLED;
}

static int xgene_dma_init_bufpool(struct xgene_dma_chan *chan,
				  struct xgene_qmtm_qinfo *bufpool)
{
	dma_addr_t buf_addr;
	void *desc16;
	void *buf;
	int i;

	for (i = 0; i < bufpool->count; i++) {
		desc16 = &bufpool->msg16[i];
		memset(desc16, 0, sizeof(struct xgene_qmtm_msg16));
		buf = devm_kzalloc(chan->dev,
				   DMA_MAX_BYTE_CNT, GFP_KERNEL);
		if (!buf)
			return -ENOMEM;

		buf_addr = dma_map_single(chan->dev, buf,
					  DMA_MAX_BYTE_CNT, DMA_TO_DEVICE);

		DMA_DESC_C_SET(desc16); /* Coherent IO */
		DMA_DESC_IN_SET(desc16);
		DMA_DESC_BUFADDR_SET(desc16 + 8, buf_addr);
		DMA_DESC_BUFLEN_SET(desc16 + 8, DMA_16K_BUFFER_LEN_CODE);
		DMA_DESC_FPQ_NUM_SET(desc16, QMTM_QUEUE_ID(DMA_QMTM_IP,
							   bufpool->queue_id));
		DMA_DESC_RTYPE_SET(desc16, DMA_RING_OWNER);

		xgene_dma_cpu_to_le64(desc16, 2);
	}

	iowrite32(bufpool->count, bufpool->command);

	return 0;
}

static int xgene_dma_allocate_ring(struct xgene_dma_chan *chan,
				   struct xgene_qmtm_qinfo *ring,
				   enum xgene_qmtm_qtype qtype, u8 slave)
{
	int ret;

	memset(ring, 0, sizeof(*ring));
	ring->qtype = qtype;
	ring->slave = slave;
	ring->qsize = QSIZE_64KB;
	ring->qaccess = QACCESS_ALT;
	ring->flags = XGENE_SLAVE_DEFAULT_FLAGS;

	ret = xgene_qmtm_set_qinfo(ring);
	if (ret != QMTM_OK) {
		chan_err(chan, "Failed to allocate ring\n");
		return ret;
	}

	chan_dbg(chan, "Queue id %d PBN %d desc 0x%p IRQ %d\n",
		 ring->queue_id, ring->pbn, ring->msg32, ring->irq);

	return ret;
}

static int xgene_dma_init_rings(struct xgene_dma *pdma)
{
	struct xgene_dma_chan *chan;
	u8 slave_i = pdma->sdev->slave_i;
	u8 slave = pdma->sdev->slave;
	int i, ret;

	for (i = 0; i < DMA_MAX_CHANNEL; i++) {
		chan = &pdma->chan[i];

		/* Allocate Tx ring descriptor */
		ret = xgene_dma_allocate_ring(chan, &chan->tx_ring,
					      QTYPE_PQ, slave);
		if (ret != QMTM_OK)
			return ret;

		/* Allocate Rx ring descriptor */
		ret = xgene_dma_allocate_ring(chan, &chan->rx_ring,
					      QTYPE_PQ, slave_i);
		if (ret != QMTM_OK)
			return ret;

		/* Set Rx irq and max outstanding request to channel */
		chan->rx_irq = chan->rx_ring.irq;
		chan->max_outstanding = chan->rx_ring.count;

		if (chan->id == DMA_FLYBY_CHANNEL) {
			/* Allocate buffer pool ring descriptor */
			ret = xgene_dma_allocate_ring(chan, &chan->bufpool,
						      QTYPE_FP, slave);
			if (ret != QMTM_OK)
				return ret;

			/* Initialize buffer pool*/
			ret = xgene_dma_init_bufpool(chan, &chan->bufpool);
			if (ret) {
				chan_err(chan, "Failed to Initialize buffer pool\n");
				return ret;
			}
		}
	}

	return ret;
}

static void xgene_dma_delete_rings(struct xgene_dma *pdma)
{
	struct xgene_dma_chan *chan;
	int i;

	for (i = 0; i < DMA_MAX_CHANNEL; i++) {
		chan = &pdma->chan[i];

		xgene_qmtm_clr_qinfo(&chan->tx_ring);
		xgene_qmtm_clr_qinfo(&chan->rx_ring);

		if (i == DMA_FLYBY_CHANNEL)
			xgene_qmtm_clr_qinfo(&chan->bufpool);
	}
}

static void xgene_dma_enable(struct xgene_dma *pdma)
{
	u32 val;

	/* Configure and enable DMA engine */
	val = ioread32(pdma->csr_dma + DMA_GCR);
	DMA_CH_SETUP(val);
	DMA_ENABLE(val);
	iowrite32(val, pdma->csr_dma + DMA_GCR);
}

static void xgene_dma_disable(struct xgene_dma *pdma)
{
	u32 val;

	val = ioread32(pdma->csr_dma + DMA_GCR);
	DMA_DISABLE(val);
	iowrite32(val, pdma->csr_dma + DMA_GCR);
}

static void xgene_dma_mask_interrupts(struct xgene_dma *pdma)
{
	/*
	 * Mask DMA ring overflow, underflow and
	 * AXI write/read error interrupts
	 */
	iowrite32(DMA_INT_ALL_MASK, pdma->csr_dma + DMA_RING_INT0_MASK);
	iowrite32(DMA_INT_ALL_MASK, pdma->csr_dma + DMA_RING_INT1_MASK);
	iowrite32(DMA_INT_ALL_MASK, pdma->csr_dma + DMA_RING_INT2_MASK);
	iowrite32(DMA_INT_ALL_MASK, pdma->csr_dma + DMA_RING_INT3_MASK);
	iowrite32(DMA_INT_ALL_MASK, pdma->csr_dma + DMA_RING_INT4_MASK);

	/* Mask DMA error interrupts */
	iowrite32(DMA_INT_ALL_MASK, pdma->csr_dma + DMA_INT_MASK);
}

static void xgene_dma_unmask_interrupts(struct xgene_dma *pdma)
{
	/*
	 * Unmask DMA ring overflow, underflow and
	 * AXI write/read error interrupts
	 */
	iowrite32(DMA_INT_ALL_UNMASK, pdma->csr_dma + DMA_RING_INT0_MASK);
	iowrite32(DMA_INT_ALL_UNMASK, pdma->csr_dma + DMA_RING_INT1_MASK);
	iowrite32(DMA_INT_ALL_UNMASK, pdma->csr_dma + DMA_RING_INT2_MASK);
	iowrite32(DMA_INT_ALL_UNMASK, pdma->csr_dma + DMA_RING_INT3_MASK);
	iowrite32(DMA_INT_ALL_UNMASK, pdma->csr_dma + DMA_RING_INT4_MASK);

	/* Unmask DMA error interrupts */
	iowrite32(DMA_INT_ALL_UNMASK, pdma->csr_dma + DMA_INT_MASK);
}

static void xgene_dma_init_hw(struct xgene_dma *pdma)
{
	u32 val;

	/* Associate DMA ring to corresponding ring HW */
	iowrite32(DMA_ASSOC_RING_MNGR1,
		  pdma->csr_dma + DMA_CFG_RING_WQ_ASSOC);
	iowrite32(DMA_ASSOC_RING_MNGR1,
		  pdma->csr_dma + DMA_CFG_RING_FQ_ASSOC);

	/* Configure RAID6 polynomial control setting */
	if (is_pq_enabled(pdma))
		iowrite32(DMA_RAID6_MULTI_CTRL(0x1D),
			  pdma->csr_dma + DMA_RAID6_CONT);
	else
		dev_info(pdma->dev, "PQ is disabled in HW\n");

	xgene_dma_enable(pdma);
	xgene_dma_unmask_interrupts(pdma);

	/* Get DMA id and version info */
	val = ioread32(pdma->csr_dma + DMA_IPBRR);

	/* DMA device info */
	dev_info(pdma->dev,
		 "X-Gene DMA v%d.%02d.%02d driver registered %d channels",
		 DMA_REV_NO_RD(val), DMA_BUS_ID_RD(val),
		 DMA_DEV_ID_RD(val), DMA_MAX_CHANNEL);
}

static int xgene_dma_init_mem(struct xgene_dma *pdma)
{
	/* Bring up memory */
	iowrite32(0x0, pdma->csr_dma + DMA_MEM_RAM_SHUTDOWN);

	/* Force a barrier */
	ioread32(pdma->csr_dma + DMA_MEM_RAM_SHUTDOWN);

	/* reset may take up to 1ms */
	usleep_range(1000, 1100);

	if (ioread32(pdma->csr_dma + DMA_BLK_MEM_RDY)
		!= DMA_BLK_MEM_RDY_VAL) {
		dev_err(pdma->dev,
			"Failed to release DMA memory from shutdown\n");
		return -ENODEV;
	}

	return 0;
}

static int xgene_dma_request_irqs(struct xgene_dma *pdma)
{
	struct xgene_dma_chan *chan;
	int ret, i, j;

	/* Register DMA error irq */
	ret = devm_request_irq(pdma->dev, pdma->err_irq, xgene_dma_err_isr,
			       0, "dma_error", pdma);
	if (ret) {
		dev_err(pdma->dev,
			"Failed to register error IRQ %d\n", pdma->err_irq);
		return ret;
	}

	/* Register DMA channel rx irq */
	for (i = 0; i < DMA_MAX_CHANNEL; i++) {
		chan = &pdma->chan[i];
		ret = devm_request_irq(chan->dev, chan->rx_irq,
				       xgene_dma_chan_ring_isr,
				       0, chan->name, chan);
		if (ret) {
			chan_err(chan, "Failed to register Rx IRQ %d\n",
				 chan->rx_irq);
			devm_free_irq(pdma->dev, pdma->err_irq, pdma);

			for (j = 0; j < i; j++) {
				chan = &pdma->chan[i];
				devm_free_irq(chan->dev, chan->rx_irq, chan);
			}

			return ret;
		}
	}

	return 0;
}

static void xgene_dma_free_irqs(struct xgene_dma *pdma)
{
	struct xgene_dma_chan *chan;
	int i;

	/* Free DMA device error irq */
	devm_free_irq(pdma->dev, pdma->err_irq, pdma);

	for (i = 0; i < DMA_MAX_CHANNEL; i++) {
		chan = &pdma->chan[i];
		devm_free_irq(chan->dev, chan->rx_irq, chan);
	}
}

static void xgene_dma_set_caps(struct xgene_dma_chan *chan,
			       struct dma_device *dma_dev)
{
	/* Initialize DMA device capability mask */
	dma_cap_zero(dma_dev->cap_mask);

	/* Set DMA device capability */
	if (xgene_dma_is_memcpy_enabled(chan))
		dma_cap_set(DMA_MEMCPY, dma_dev->cap_mask);

	if (xgene_dma_is_sg_enabled(chan))
		dma_cap_set(DMA_SG, dma_dev->cap_mask);

	/* Basically here, the X-Gene SoC DMA engine channel 0 supports XOR
	 * and channel 1 supports XOR, PQ both. First thing here is we have
	 * mechanism in hw to enable/disable PQ/XOR supports on channel 1,
	 * we can make sure this by reading SoC Efuse register.
	 * Second thing, we have hw errata that if we run channel 0 and
	 * channel 1 simultaneously with executing XOR and PQ request,
	 * suddenly DMA engine hangs, So here we enable XOR on channel 0 only
	 * if XOR and PQ supports on channel 1 is disabled.
	 */
	if (xgene_dma_is_xor_enabled(chan))
		dma_cap_set(DMA_XOR, dma_dev->cap_mask);

	if (xgene_dma_is_pq_enabled(chan))
		dma_cap_set(DMA_PQ, dma_dev->cap_mask);

	/* Set base and prep routines */
	dma_dev->dev = chan->dev;
	dma_dev->device_alloc_chan_resources = xgene_dma_alloc_chan_resources;
	dma_dev->device_free_chan_resources = xgene_dma_free_chan_resources;
	dma_dev->device_issue_pending = xgene_dma_issue_pending;
	dma_dev->device_tx_status = xgene_dma_tx_status;

	if (dma_has_cap(DMA_MEMCPY, dma_dev->cap_mask))
		dma_dev->device_prep_dma_memcpy = xgene_dma_prep_memcpy;

	if (dma_has_cap(DMA_SG, dma_dev->cap_mask))
		dma_dev->device_prep_dma_sg = xgene_dma_prep_sg;

	if (dma_has_cap(DMA_XOR, dma_dev->cap_mask)) {
		dma_dev->device_prep_dma_xor = xgene_dma_prep_xor;
		dma_dev->max_xor = DMA_MAX_XOR_SRC;
		dma_dev->xor_align = DMA_XOR_ALIGNMENT;
	}

	if (dma_has_cap(DMA_PQ, dma_dev->cap_mask)) {
		dma_dev->device_prep_dma_pq = xgene_dma_prep_pq;
		dma_dev->max_pq = DMA_MAX_XOR_SRC;
		dma_dev->pq_align = DMA_XOR_ALIGNMENT;
	}
}

static int xgene_dma_async_register(struct xgene_dma *pdma, int id)
{
	struct xgene_dma_chan *chan = &pdma->chan[id];
	struct dma_device *dma_dev = &pdma->dma_dev[id];
	int ret;

	chan->dma_chan.device = dma_dev;

	spin_lock_init(&chan->lock);
	INIT_LIST_HEAD(&chan->ld_pending);
	INIT_LIST_HEAD(&chan->ld_running);
	INIT_LIST_HEAD(&chan->ld_completed);
	tasklet_init(&chan->tasklet, xgene_dma_tasklet_cb,
		     (unsigned long)chan);

	chan->pending = 0;
	chan->desc_pool = NULL;
	dma_cookie_init(&chan->dma_chan);

	/* Setup dma device capabilities and prep routines */
	xgene_dma_set_caps(chan, dma_dev);

	/* Initialize DMA device list head */
	INIT_LIST_HEAD(&dma_dev->channels);
	list_add_tail(&chan->dma_chan.device_node, &dma_dev->channels);

	/* Register with Linux async DMA framework*/
	ret = dma_async_device_register(dma_dev);
	if (ret) {
		chan_err(chan, "Failed to register async device %d", ret);
		tasklet_kill(&chan->tasklet);

		return ret;
	}

	/* DMA capability info */
	dev_info(pdma->dev,
		 "%s: CAPABILITY ( %s%s%s%s%s)\n",
		 dma_chan_name(&chan->dma_chan),
		 dma_has_cap(DMA_MEMCPY, dma_dev->cap_mask) ? "MEMCPY " : "",
		 dma_has_cap(DMA_SG, dma_dev->cap_mask) ? "SGCPY " : "",
		 dma_has_cap(DMA_XOR, dma_dev->cap_mask) ? "XOR " : "",
		 dma_has_cap(DMA_PQ, dma_dev->cap_mask) ? "PQ " : "",
		 (chan->id == DMA_FLYBY_CHANNEL) ? "iSCSI-CRC32c " : "");

	return 0;
}

static int xgene_dma_init_async(struct xgene_dma *pdma)
{
	int ret, i, j;

	for (i = 0; i < DMA_MAX_CHANNEL ; i++) {
		ret = xgene_dma_async_register(pdma, i);
		if (ret) {
			for (j = 0; j < i; j++) {
				dma_async_device_unregister(&pdma->dma_dev[j]);
				tasklet_kill(&pdma->chan[j].tasklet);
			}

			return ret;
		}
	}

	return ret;
}

static void xgene_dma_async_unregister(struct xgene_dma *pdma)
{
	int i;

	for (i = 0; i < DMA_MAX_CHANNEL; i++)
		dma_async_device_unregister(&pdma->dma_dev[i]);
}

static void xgene_dma_init_channels(struct xgene_dma *pdma)
{
	struct xgene_dma_chan *chan;
	int i;

	for (i = 0; i < DMA_MAX_CHANNEL; i++) {
		chan = &pdma->chan[i];
		chan->dev = pdma->dev;
		chan->pdma = pdma;
		chan->id = i;
		sprintf(chan->name, "dmachan%d", chan->id);
	}
}

static int xgene_dma_get_resources(struct platform_device *pdev,
				   struct xgene_dma *pdma)
{
	struct resource *res;

	/* Get DMA csr region */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&pdev->dev, "Failed to get csr region\n");
		return -ENXIO;
	}

	pdma->csr_dma = devm_ioremap(&pdev->dev, res->start,
				     resource_size(res));
	if (IS_ERR(pdma->csr_dma)) {
		dev_err(&pdev->dev, "Failed to ioremap csr region");
		return PTR_ERR(pdma->csr_dma);
	}

	/* Get efuse csr region */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 1);
	if (!res) {
		dev_err(&pdev->dev, "Failed to get efuse csr region\n");
		return -ENXIO;
	}

	pdma->csr_efuse = devm_ioremap(&pdev->dev, res->start,
				       resource_size(res));
	if (IS_ERR(pdma->csr_efuse)) {
		dev_err(&pdev->dev, "Failed to ioremap efuse csr region");
		return PTR_ERR(pdma->csr_efuse);
	}

	/* Get DMA error interrupt */
	pdma->err_irq = platform_get_irq(pdev, 0);
	if (pdma->err_irq <= 0) {
		dev_err(&pdev->dev, "Failed to get Error IRQ\n");
		return -ENXIO;
	}

	return 0;
}

static int xgene_dma_runtime_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct xgene_dma *pdma = platform_get_drvdata(pdev);

	if (!IS_ERR(pdma->clk))
		clk_disable_unprepare(pdma->clk);

	return 0;
}

static int xgene_dma_runtime_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct xgene_dma *pdma = platform_get_drvdata(pdev);
	int ret;

	if (!IS_ERR(pdma->clk)) {
		ret = clk_prepare_enable(pdma->clk);
		if (ret) {
			dev_err(dev, "Failed to enable clk %d\n", ret);
			return ret;
		}
	}

	return 0;
}

static int xgene_dma_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node;
	const char default_name[] = "PKTDMA";
	struct xgene_dma *pdma;
	const char *name;
	u8 wq_pbn_start;
	u8 wq_pbn_count;
	u8 fq_pbn_start;
	u8 fq_pbn_count;
	u32 qmtm_ip;
	u32 info[5];
	int ret;

	pdma = devm_kzalloc(&pdev->dev, sizeof(*pdma), GFP_KERNEL);
	if (!pdma)
		return -ENOMEM;

	pdata = pdma;
	pdma->dev = &pdev->dev;
	platform_set_drvdata(pdev, pdma);

	ret = xgene_dma_get_resources(pdev, pdma);
	if (ret)
		return ret;

	/* EFI BOOT and ACPI */
	if (efi_enabled(EFI_BOOT)) {
		name = &default_name[0];
		qmtm_ip = 0x1;
		wq_pbn_start = 0x0;
		wq_pbn_count = 4;
		fq_pbn_start = 0x20;
		fq_pbn_count = 8;
	} else {
		/* Fetch the slave name from DTS */
		ret = of_property_read_string(np, "slave_name", &name);
		if (ret) {
			dev_err(&pdev->dev, "slave_name is missing in DTS\n");
			return ret;
		}

		/* Fetch slave info from DTS */
		ret = of_property_read_u32_array(np, "slave_info",
						 info, ARRAY_SIZE(info));
		if (ret) {
			dev_err(&pdev->dev, "slave_info is missing in DTS\n");
			return ret;
		}

		qmtm_ip = info[0];
		wq_pbn_start = info[1];
		wq_pbn_count = info[2];
		fq_pbn_start = info[3];
		fq_pbn_count = info[4];
	}

	pdma->sdev = xgene_qmtm_set_sdev(name, qmtm_ip, wq_pbn_start,
					 wq_pbn_count, fq_pbn_start,
					 fq_pbn_count);
	if (!pdma->sdev) {
		dev_err(&pdev->dev, "QMTM%d slave %s error\n", qmtm_ip, name);
		return -ENODEV;
	}

	pdma->clk = devm_clk_get(&pdev->dev, NULL);
	if (IS_ERR(pdma->clk) && !efi_enabled(EFI_BOOT)) {
		dev_err(&pdev->dev, "Failed to get clk\n");
		return PTR_ERR(pdma->clk);
	}

	/* Setup runtime resume and suspend */
	pm_runtime_enable(&pdev->dev);
	if (!pm_runtime_enabled(&pdev->dev)) {
		ret = xgene_dma_runtime_resume(&pdev->dev);
		if (ret) {
			dev_err(&pdev->dev,
				"Failed to resume at runtime %d\n", ret);
			goto err_rt_resume;
		}
	}

	/* Enable clk before accessing registers */
	if (!IS_ERR(pdma->clk)) {
		ret = clk_prepare_enable(pdma->clk);
		if (ret) {
			dev_err(&pdev->dev, "Failed to enable clk %d\n", ret);
			goto err_pm_enable;
		}
	}

	/* Remove DMA RAM out of shutdown */
	ret = xgene_dma_init_mem(pdma);
	if (ret)
		goto err_pm_enable;

	/* Set DMA and coherent mask */
	pdev->dev.dma_mask = &pdev->dev.coherent_dma_mask;
	pdev->dev.coherent_dma_mask = DMA_BIT_MASK(42);

	/* Initialize DMA channels software state */
	xgene_dma_init_channels(pdma);

	/* Configue DMA rings */
	ret = xgene_dma_init_rings(pdma);
	if (ret)
		goto err_pm_enable;

	ret = xgene_dma_request_irqs(pdma);
	if (ret)
		goto err_request_irq;

	/* Configure and enable DMA engine */
	xgene_dma_init_hw(pdma);

	/* Register DMA device with linux async framework */
	ret = xgene_dma_init_async(pdma);
	if (ret)
		goto err_async_init;

	return 0;

err_async_init:
	xgene_dma_free_irqs(pdma);

err_request_irq:
	xgene_dma_delete_rings(pdma);

err_pm_enable:
	if (!pm_runtime_status_suspended(&pdev->dev))
		xgene_dma_runtime_suspend(&pdev->dev);

err_rt_resume:
	pm_runtime_disable(&pdev->dev);
	xgene_qmtm_clr_sdev(pdma->sdev);
	return ret;
}

static int xgene_dma_remove(struct platform_device *pdev)
{
	struct xgene_dma *pdma = platform_get_drvdata(pdev);
	struct xgene_dma_chan *chan;
	int i;

	xgene_dma_async_unregister(pdma);

	/* Mask interrupts and disable DMA engine */
	xgene_dma_mask_interrupts(pdma);
	xgene_dma_disable(pdma);
	xgene_dma_free_irqs(pdma);

	for (i = 0; i < DMA_MAX_CHANNEL; i++) {
		chan = &pdma->chan[i];
		tasklet_kill(&chan->tasklet);
	}

	pm_runtime_disable(&pdev->dev);
	if (!pm_runtime_status_suspended(&pdev->dev))
		xgene_dma_runtime_suspend(&pdev->dev);

	/* Delete rings allocated to this DMA device */
	xgene_dma_delete_rings(pdma);
	xgene_qmtm_clr_sdev(pdma->sdev);

	return 0;
}

static const struct dev_pm_ops xgene_dma_pm_ops = {
#ifdef CONFIG_PM_RUNTIME
	.runtime_suspend = xgene_dma_runtime_suspend,
	.runtime_resume = xgene_dma_runtime_resume,
#endif
};

#ifdef CONFIG_ACPI
static const struct acpi_device_id xgene_dma_acpi_match[] = {
	{"APMC0D16", },
	{},
};
MODULE_DEVICE_TABLE(acpi, xgene_dma_acpi_match);
#endif

static const struct of_device_id xgene_dma_of_match_ptr[] = {
	{.compatible = "apm,xgene-pktdma",},
	{},
};
MODULE_DEVICE_TABLE(of, xgene_dma_of_match_ptr);

static struct platform_driver xgene_dma_driver = {
	.probe = xgene_dma_probe,
	.remove = xgene_dma_remove,
	.driver = {
		.name = "X-Gene-PktDMA",
		.owner = THIS_MODULE,
		.pm = &xgene_dma_pm_ops,
		.of_match_table = xgene_dma_of_match_ptr,
#ifdef CONFIG_ACPI
		.acpi_match_table = ACPI_PTR(xgene_dma_acpi_match),
#endif
	},
};

static int __init xgene_dma_init(void)
{
	return platform_driver_register(&xgene_dma_driver);
}
late_initcall(xgene_dma_init);

static void __exit xgene_dma_exit(void)
{
	platform_driver_unregister(&xgene_dma_driver);
}
module_exit(xgene_dma_exit);

MODULE_DESCRIPTION("APM X-Gene SoC DMA driver");
MODULE_AUTHOR("Rameshwar Prasad Sahu <rsahu@apm.com>");
MODULE_AUTHOR("Loc Ho <lho@apm.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("1.0");
