/*
 * PCIE LOOPBACK TEST
 *
 * Copyright 2012 AMCC
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/cacheflush.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/of_device.h>
#include <linux/dmaengine.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/mm.h>
#include <linux/crc16.h>
#include <linux/random.h>
#include <linux/dma/xgene-pcie-dmalib.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>

#define NUM_PCI_RES		4
#define MAX_DMA_CHAN		4
#define PCIE_TEST		"X-Gene EP"
#define PCIE_TEST_VERSION	"0.1"
#define PATTERN			0xdeadbeef
#define OFFSET			0x4

#define BUF_LEN			PAGE_SIZE * 0x400
#define TOTAL_SIZE PAGE_SIZE * 0x400 /* 4M */
#define SRC_SG_ENTS 2
#define SRC_ELEM_SIZE (TOTAL_SIZE / SRC_SG_ENTS)
#define DST_SG_ENTS 4
#define DST_ELEM_SIZE (TOTAL_SIZE / DST_SG_ENTS)

enum {
	NORMAL_TEST = 0,
	MEMCPY_TEST,
	GATHER_TEST,
	SCATTER_GATHER_TEST,
	EP_DMA_WRITE,
	EP_DMA_READ,
	EP_DMA_VERIFY,
};

int DMA_DISABLED;
int timer_expired = 0;
int test_type	= SCATTER_GATHER_TEST;
int dma_test	= DMA_BIDIRECTIONAL;
int itr_cnt	= 1;
int duration	= 1;
int sg_len	= 8192;

module_param_named(dma, test_type, int, S_IRUGO);
MODULE_PARM_DESC(dma, "DMA Test type - 0 : NORMAL, 1 : DMA, 2 : SG DMA"
		"(Default : 0 (Normal))");
module_param_named(test, dma_test, int, S_IRUGO);
MODULE_PARM_DESC(test,  "DMA Test direction - 1 : to_device,"
		" 2 : from_device, 0 :both (Default: both)");
module_param_named(duration, duration, int, S_IRUGO);
MODULE_PARM_DESC(duration, "DMA test duration (Default: 1sec)");

extern void xgene_pcie_gen_msi (int msg);
extern void xgene_pciedma_hw_init(void);
extern void xgene_pcievent_init_wait(void);
extern irqreturn_t xgene_event_isr(int irq, void *data);
extern int xgene_dma_get_compl_count(void);
extern void xgene_dma_set_compl_count(int count);
static struct xgene_dma_completion dma_completion;
struct bar_info {
	int	enabled;
	void	*vaddr;
	void	*paddr;
	u32	len;
};

struct pci_ep_info {
	struct pci_dev	*dev;
	int		irq;
	struct bar_info	bars[NUM_PCI_RES];
	struct dma_chan *chan[MAX_DMA_CHAN];
	struct proc_dir_entry *dma_proc;
};


static void memcpy_completion(void *param);
dma_cookie_t
dma_async_pcie_memcpy_buf_to_buf(struct dma_chan *chan, u64 dest, u64 src,
		size_t len, enum dma_data_direction dir,
		void *callback_func, void *callback_param);

static void timer_callback(unsigned long data)
{
	timer_expired = 1;
}

static void init_test_timer(struct timer_list *test_timer, int time_sec)
{
	init_timer(test_timer);
	test_timer->expires = jiffies + time_sec*HZ;
	test_timer->data = 0;
	test_timer->function = timer_callback;
	add_timer(test_timer);
}

static int ob_perf_test(struct pci_ep_info *ep_info, void *addr, int len, int dir, int type)
{
	struct timer_list test_timer;
	struct timeval starttime, endtime;
	u64 temp = 0, total_time, total_len = 0;
	u64 src, dest;
	u64 pci_addr = (u64)ep_info->bars[0].paddr & 0xffffffff;
	u64 count = 0;

	dev_info (&ep_info->dev->dev, "Starting PCIe outbound %d sec %s %s test\n", duration,
			(dir == 1) ? "write" : "read", (type == 1) ? "DMA": "Normal");
	timer_expired = 0;
	init_test_timer(&test_timer, duration);
	do_gettimeofday(&starttime);
	if (type == 1) {
		if (dir == DMA_FROM_DEVICE) {
			src = pci_addr;
			dest = (u64) addr;
		} else {
			dest = pci_addr;
			src = (u64) addr;
		}
		while(1) {
			atomic_set(&dma_completion.flag, 0);
			dma_async_pcie_memcpy_buf_to_buf(ep_info->chan[0],
					dest, src, len, dir, memcpy_completion, &dma_completion);
			if (!wait_event_timeout(dma_completion.cmpl,
						atomic_read(&dma_completion.flag),100))
				pr_err("%s: timeout\n", __func__);
			count++;
			total_len += len;
			if (timer_expired) {
				del_timer(&test_timer);
				break;
			}
		}
	} else {
		if (ep_info->bars[0].enabled) {
			while (1) {
				u64 *address = addr;
				if (dir == DMA_FROM_DEVICE)
					*address = temp;
				else
					temp = *address;
				total_len += len;
				if (timer_expired) {
					del_timer(&test_timer);
					break;
				}
			}
		}
	}
	do_gettimeofday(&endtime);
	total_time = endtime.tv_sec - starttime.tv_sec;
	total_len /= (1024 * 1024);

	pr_info("Performance with %d bytes per transaction\n", len);
	pr_info("Total req sent      : %llu\n", count);
	pr_info("Total data transfer : %llu MB\n", total_len);
	pr_info("Total duration      : %llu sec\n", total_time);
	total_len /= total_time;
	total_len *= 8;
	pr_info("Throughput          : %llu Mbps\n", total_len);
	return 0;
}


static ssize_t xgene_pcie_dma_write_irq(struct file *file, const char __user *buf,
		size_t count, loff_t *data)
{
	xgene_pcie_gen_msi(0);
	return count;
}

static void xgene_pcie_dma_test_start(struct pci_ep_info *);

static ssize_t xgene_pcie_dma_write_sg_len(struct file *file, const char __user *buf,
		size_t count, loff_t *data)
{
	static char temp_buf[5];
	unsigned long temp;
	const char *str;

	if (copy_from_user(temp_buf, buf, count))
		count = -EFAULT;

	temp_buf[count] = 0;
	str = &temp_buf[0];
	temp = simple_strtoul(str, NULL, 0);
	pr_debug("%s: sg_len = %lu\n", __func__, temp);
	if (temp <= 0)
		return -EINVAL;
	sg_len = temp;
	return count;
}

static ssize_t xgene_pcie_dma_write_duration(struct file *file, const char __user *buf,
		size_t count, loff_t *data)
{
	static char temp_buf[5];
	unsigned long temp;
	const char *str;

	if (copy_from_user(temp_buf, buf, count))
		count = -EFAULT;

	temp_buf[count] = 0;
	str = &temp_buf[0];
	temp = simple_strtoul(str, NULL, 0);
	pr_debug("%s: duration = %lu\n", __func__, temp);
	if (temp <= 0)
		return -EINVAL;
	duration = temp;
	return count;
}

static ssize_t xgene_pcie_dma_write_proc(struct file *file, const char __user *buf,
		size_t count, loff_t *data)
{
	struct pci_ep_info *ep_info = (struct pci_ep_info *)PDE_DATA(file_inode(file));
	static char dma_test_buf[2] = {'\0'};

	BUG_ON(!ep_info);

	if (count != 2) {
		count = -EINVAL;
		goto usage;
	}

	if (copy_from_user(dma_test_buf, buf, sizeof(dma_test_buf) - 1)) 
		return -EFAULT;

	test_type = dma_test_buf[0] - '0';
	if (test_type != NORMAL_TEST && DMA_DISABLED) {
		pr_err("DMA Tests are disabled");
		count = -EINVAL;
		goto usage;
	}

	if (test_type < NORMAL_TEST || test_type > EP_DMA_VERIFY) {
		pr_err("Invalid DMA Test Type\n");
		count = -EINVAL;
		goto usage;
	}
	xgene_pcie_dma_test_start(ep_info);
	return count;
usage:
	pr_info( "usage:\n");
	pr_info( "echo \"<test_type>\" > "
			" /proc/xgene_pcie_test/test-config\n");
	pr_info( "Test Type - 0 : Outbound, 1 : Memcpy, 2 : Gather, 3 : Scatter-Gather\n");
	pr_info( "Example:\n");
	pr_info( "echo '0' > /proc/xgene_pcie_test/test-config\n");
	return count;
}

static const struct file_operations test_config_fops = {
	.owner = THIS_MODULE,
	.write = xgene_pcie_dma_write_proc,
};

static const struct file_operations sg_len_fops = {
	.owner = THIS_MODULE,
	.write = xgene_pcie_dma_write_sg_len,
};

static const struct file_operations duration_fops = {
	.owner = THIS_MODULE,
	.write = xgene_pcie_dma_write_duration,
};

static const struct file_operations irq_fops = {
	.owner = THIS_MODULE,
	.write = xgene_pcie_dma_write_irq,
};

#ifdef CONFIG_XGENE_PCIE_DMA
static int init_dma_proc(struct pci_ep_info *ep_info)
{
	struct proc_dir_entry *entry;

	ep_info->dma_proc = proc_mkdir("xgene_pcie_test", NULL);
	if (ep_info->dma_proc == NULL) {
		pr_err("/proc/xgene_pcie_test creation failed\n");
		return -1;
	}

	entry = proc_create_data("test-config", S_IRWXUGO, ep_info->dma_proc,
			&test_config_fops, ep_info);
	pr_info("/proc/xgene_pcie_test/test-config created!\n");
	if(!DMA_DISABLED) {
		entry = proc_create("sg_len", S_IRWXUGO, ep_info->dma_proc,
				&sg_len_fops);
		pr_debug("/proc/xgene_pcie_test/sg_len created!\n");
		entry = proc_create("duration", S_IRWXUGO, ep_info->dma_proc,
				&duration_fops);
		pr_debug("/proc/xgene_pcie_test/duration created!\n");
	}
	entry = proc_create_data("irq", S_IRWXUGO, ep_info->dma_proc,
			&irq_fops, ep_info);
	pr_info("/proc/xgene_pcie_test/irq created!\n");
	return 0;
}

static void sg_dma_done(void *param)
{
	struct xgene_dma_completion *sg_dma = param;
	struct dma_slave_config *dma_slave = sg_dma->slave;
	struct scatterlist *sg, *sgl;
	struct page *pages;
	int sg_cnt = sg_dma->sg_cnt;
	int total_len = 0;
	int direction;
	int i;

	sgl = sg_dma->sgl;
	pages = (struct page *) sgl->page_link;
	direction = (dma_slave->direction == DMA_MEM_TO_DEV) ? DMA_TO_DEVICE :
		DMA_FROM_DEVICE;
	for_each_sg(sgl, sg, sg_cnt, i) {
		if (direction == DMA_TO_DEVICE) {
			dma_unmap_page(sg_dma->device->dev, sg_dma_address(sg),
					sg_dma_len(sg), direction);
		}
		total_len += sg_dma_len(sg);
	}
	if (direction == DMA_TO_DEVICE)
		__free_pages(pages, get_order(total_len));
	kfree(sgl);
	kfree(dma_slave);
	atomic_set(&sg_dma->flag, 1);
	wake_up(&sg_dma->cmpl);
	pr_info("SG DMA Done\n");
}
#endif
struct xgene_dma_completion *prep_sg_dma(struct dma_device *device, int dir)
{
	struct scatterlist *sg, *sgl;
	struct xgene_dma_completion *sg_dma;
	struct page *pages = NULL;
	dma_addr_t addr = 0;
	int sg_cnt = sg_len / PAGE_SIZE;
	int len;
	int consumed = 0;
	int i;

	sg_dma = kzalloc(sizeof(*sg_dma), GFP_KERNEL);
	if (!sg_dma)
		goto fail_sg;

	if (sg_len % PAGE_SIZE)
		sg_cnt++;

	pr_debug("SG DMA - sg_len : %d, sg_cnt = %d\n", sg_len, sg_cnt);
	sgl = kzalloc(sizeof(*sgl)*sg_cnt, GFP_KERNEL);
	if (!sgl)
		goto fail_sg_dma;

	sg_init_table(sgl, sg_cnt);

	if (dir == DMA_TO_DEVICE) {
		pages = alloc_pages(GFP_KERNEL, get_order(sg_len));
		if (!pages)
			goto fail_page;
	}

	len = (sg_len > PAGE_SIZE) ? PAGE_SIZE : sg_len;
	for_each_sg(sgl, sg, sg_cnt, i) {
		if ((consumed + PAGE_SIZE) > sg_len)
			len = sg_len - consumed;
		if (dir == DMA_TO_DEVICE) {
			memset(page_address(pages), i + 1, len);
			sg_set_page(sg, pages, len, 0);
			addr = dma_map_page(device->dev, pages, 0, len, dir);
			BUG_ON(!addr);
			sg_dma_address(sg) = addr;
			pages++;
		} else {
			sg_dma_address(sg) = addr;
			addr += len;
		}
		sg_dma_len(sg) = len;
		consumed += len;
		pr_info("Scatterlist %d: Addr - 0x%08llx, Len - %d\n",
				i, sg_dma_address(sg), sg_dma_len(sg));
	}

	sg_dma->sgl = sgl;
	sg_dma->sg_cnt = sg_cnt;

	return sg_dma;
fail_page:
	kfree(sgl);
fail_sg:
	kfree(sg_dma);
fail_sg_dma:
	return NULL;
}

#ifdef CONFIG_XGENE_PCIE_DMA
static int do_pcie_dma_sg(struct pci_ep_info *ep_info)
{
	int j, k, ret = 0, ch;
	struct dma_addr_map *srcmap[MAX_DMA_CHAN], *dstmap[MAX_DMA_CHAN];
	struct device *dev = &ep_info->dev->dev;
	dma_addr_t pci_addr = (u64)ep_info->bars[2].paddr & 0xffffffff;
	struct timeval starttime, endtime;
	u64 total_time;

	for (k = 0; k < itr_cnt; k++) {
		dev_info(dev, "\nSG DMA Test - total buffer size %X : %d\n",
				(u32)TOTAL_SIZE, k);
		dev_dbg (dev, "AXI to PCIe\n");
		xgene_dma_set_compl_count(0);
		do_gettimeofday(&starttime);
		for (ch = 0; ch < MAX_DMA_CHAN; ch++) {
			dev_dbg (dev, "Channel %d\n",ch);
			srcmap[ch] = xgene_pcie_dma_sg_write(ep_info->chan[ch],
					pci_addr + (16 * XGENE_DMA_MAX_LEN), 16);
			if (!srcmap[ch]) {
				dev_err (dev, "SG DMA Write failed\n");
				ret = -EAGAIN;
				goto out;
			}
			ep_info->chan[ch]->device->device_issue_pending(ep_info->chan[ch]);
			dev_dbg (dev, "PCIe to AXI\n");
			dstmap[ch] = xgene_pcie_dma_sg_read(ep_info->chan[ch],
					pci_addr + (16 * XGENE_DMA_MAX_LEN), 16);
			if (!dstmap[ch]) {
				dev_err (dev, "SG DMA Read failed\n");
				ret = -EAGAIN;
				goto out;
			}
			ep_info->chan[ch]->device->device_issue_pending(ep_info->chan[ch]);
		}
		while (xgene_dma_get_compl_count() != MAX_DMA_CHAN * 2);
		do_gettimeofday(&endtime);
		total_time = endtime.tv_sec - starttime.tv_sec;
		dev_dbg (dev, "Transferred 512M in %lld secs\n", total_time);
		for (ch = 0; ch < MAX_DMA_CHAN; ch++) {
			if (xgene_mem_crc(srcmap[ch]) == xgene_mem_crc(dstmap[ch])) {
				dev_info(dev, "BIDIRECTIONAL SG TEST CHANNEL [%d][%d] PASS\n",ch, k);
			} else {
				dev_info(dev, "BIDIRECTIONAL SG TEST CHANNEL [%d][%d] FAIL\n",ch, k);
				ret = -EAGAIN;
			}
		}
	}
out:
	for (ch = 0; ch < MAX_DMA_CHAN; ch++) {
		struct dma_addr_map *s, *d;
		s = srcmap[ch];
		d = dstmap[ch];
		for (j = 0; s && s[j].len; j++) {
			kfree(s[j].virt);
		}
		for (j = 0; d && d[j].len; j++) {
			kfree(d[j].virt);
		}
	}
	return ret;
}

static int do_pcie_dma_gather(struct pci_ep_info *ep_info)
{
	struct dma_async_tx_descriptor *tx = NULL;
	struct xgene_dma_completion *sg_dma;
	struct dma_slave_config *dma_slave;
	struct dma_device *dma_dev = ep_info->chan[0]->device;
	u64 pci_addr = (u64)ep_info->bars[0].paddr & 0xffffffff;
	int dir;
	int ret = 0;
	int test_type = DMA_TO_DEVICE;

	dev_info(&ep_info->dev->dev,"SG DMA Test - ");

	dma_slave = kzalloc(sizeof(*dma_slave), GFP_KERNEL);
	if (!dma_slave)
		goto fail_dma_slave;

	switch(test_type) {
		case DMA_TO_DEVICE:
			dev_dbg(&ep_info->dev->dev,"AXI to PCIe\n");
			dma_slave->direction = DMA_MEM_TO_DEV;
			dma_slave->dst_addr = pci_addr;
			break;

		case DMA_FROM_DEVICE:
			dev_dbg(&ep_info->dev->dev,"PCIe to AXI\n");
			dma_slave->direction = DMA_DEV_TO_MEM;
			dma_slave->src_addr = pci_addr;
			break;

		default:
			dev_err(&ep_info->dev->dev,"Unsupported SG DMA Test\n");
			ret = -EINVAL;
			goto fail_dma_slave;
	}

	dma_dev->device_control(ep_info->chan[0], DMA_SLAVE_CONFIG,
			(unsigned long) dma_slave);

	dir = (dma_slave->direction == DMA_MEM_TO_DEV) ? DMA_TO_DEVICE :
		DMA_FROM_DEVICE;
	sg_dma = prep_sg_dma(dma_dev, dir);
	if (!sg_dma) {
		ret = -ENOMEM;
		goto fail_dma_slave;
	}

	init_waitqueue_head(&sg_dma->cmpl);
	sg_dma->slave = dma_slave;
	sg_dma->device = dma_dev;
	atomic_set(&sg_dma->flag, 0);
	tx = dmaengine_prep_slave_sg(ep_info->chan[0], sg_dma->sgl,
			sg_dma->sg_cnt,
			dma_slave->direction,
			DMA_PREP_INTERRUPT);
	BUG_ON(tx == NULL);
	tx->callback = sg_dma_done;
	tx->callback_param = sg_dma;
	dev_dbg(&ep_info->dev->dev,"Submit SG DMA TX\n");
	tx->tx_submit(tx);
	ep_info->chan[0]->device->device_issue_pending(ep_info->chan[0]);

	if(!wait_event_interruptible_timeout(sg_dma->cmpl,
				atomic_read(&sg_dma->flag),100))
		dev_err(&ep_info->dev->dev,"SG_DMA wait_event_timeout\n");

	kfree(sg_dma);
	return ret;

fail_dma_slave:
	kfree(dma_slave);
	return ret;
}
#endif
static void memcpy_completion(void *param)
{
	struct xgene_dma_completion *dma_complete = param;	
	if(dma_complete) {
		atomic_set(&dma_complete->flag, 1);
		wake_up(&dma_complete->cmpl);
	}
}

	dma_cookie_t
dma_async_pcie_memcpy_buf_to_buf(struct dma_chan *chan, u64 dest, u64 src,
		size_t len, enum dma_data_direction dir,
		void *callback_func, void *callback_param)
{
	struct dma_device *dev = chan->device;
	struct dma_async_tx_descriptor *tx;
	dma_addr_t dma_dest, dma_src;
	dma_cookie_t cookie;
	unsigned long flags;

	if (dir == DMA_TO_DEVICE) {
		dma_src = dma_map_single(dev->dev, (void *)src, len, DMA_TO_DEVICE);
		dma_dest = (dma_addr_t) dest;
		flags = DMA_COMPL_SRC_UNMAP_SINGLE;
	} else {
		dma_src = (dma_addr_t) src;
		dma_dest = dma_map_single(dev->dev, (void *)dest, len,
				DMA_FROM_DEVICE);
		flags = DMA_COMPL_DEST_UNMAP_SINGLE;
	}

	flags |= DMA_CTRL_ACK | DMA_PREP_INTERRUPT;
	tx = dev->device_prep_dma_memcpy(chan, dma_dest, dma_src, len, flags);
	if (!tx) {
		dev_dbg(dev->dev,"%s: Device prep memcpy failed\n", __func__);
		if (flags & DMA_COMPL_SRC_UNMAP_SINGLE)
			dma_unmap_single(dev->dev, dma_src, len, DMA_TO_DEVICE);
		else
			dma_unmap_single(dev->dev, dma_dest, len, DMA_FROM_DEVICE);
		return -ENOMEM;
	}

	tx->callback = callback_func;
	tx->callback_param = callback_param;

	cookie = tx->tx_submit(tx);
#if 1
	preempt_disable();
	__this_cpu_add(chan->local->bytes_transferred, len);
	__this_cpu_inc(chan->local->memcpy_count);
	preempt_enable();
#endif

	return cookie;
}

#ifdef CONFIG_XGENE_PCIE_DMA
static int do_pcie_dma_memcpy(struct pci_ep_info *ep_info)
{
	int i, ret = 0, j, ch;
	void *buf;
	void *validation_buf = NULL;
	u64 pci_addr = (u64)ep_info->bars[0].paddr & 0xffffffff;
	u32 offset;

	dev_info(&ep_info->dev->dev,"Starting PCIe DMA Test ");

	buf = kmalloc(BUF_LEN, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	if (dma_test == DMA_BIDIRECTIONAL) {
		validation_buf = kmalloc(BUF_LEN, GFP_KERNEL);
		if (!validation_buf) {
			kfree(buf);
			return -ENOMEM;
		}
	}
	for (i = 0; i < itr_cnt; i++) {
		for (ch = 0; ch < 4; ch++) {
			xgene_dma_set_compl_count(0);
			get_random_bytes(buf, BUF_LEN);
			memset(validation_buf, 0, BUF_LEN);
			dev_dbg (&ep_info->dev->dev, "Channel[%d], DMA TRANSFER[%d]",ch, i);
			switch(dma_test) {
				case DMA_TO_DEVICE:
					dev_dbg (&ep_info->dev->dev, "From AXI to PCIE\n");
					ob_perf_test(ep_info, buf, BUF_LEN, DMA_TO_DEVICE, 1);
					break;

				case DMA_FROM_DEVICE:
					dev_dbg (&ep_info->dev->dev, "From PCIE to AXI\n");
					ob_perf_test(ep_info, buf, BUF_LEN, DMA_FROM_DEVICE, 1);
					break;

				case DMA_BIDIRECTIONAL:
					dev_dbg (&ep_info->dev->dev, "BIDIRECTIONAL VALIDATION START\n");
					for (j = 0; j < 0x400; j++) {
						offset = j * PAGE_SIZE;
						if (xgene_pcie_dma_memcpy(ep_info->chan[ch],
									pci_addr + offset,
									(u64)(buf + offset),
									PAGE_SIZE, DMA_MEM_TO_DEV)) {
							dev_err (&ep_info->dev->dev, "Memcpy TO device failed\n");
						}
					}
					ep_info->chan[0]->device->device_issue_pending(ep_info->chan[ch]);
					for (j = 0; j < 0x400; j++) {
						offset = j * PAGE_SIZE;
						if (xgene_pcie_dma_memcpy(ep_info->chan[ch],
									(u64)(validation_buf + offset),
									pci_addr + offset,
									PAGE_SIZE, DMA_DEV_TO_MEM)) {
							dev_err (&ep_info->dev->dev, "Memcpy FROM device failed\n");
						}
					}
					ep_info->chan[0]->device->device_issue_pending(ep_info->chan[ch]);
					break;
				default:
					dev_err(&ep_info->dev->dev,"Incorrect Test\n");
					break;
			}
			while (xgene_dma_get_compl_count() != (j * 2));
			if (crc16(0, buf, BUF_LEN) == crc16(0, validation_buf, BUF_LEN))
				dev_info(&ep_info->dev->dev, "BIDIRECTIONAL MEMCPY CHANNEL %d TEST [%d] PASS\n",ch ,i);
			else
				dev_info(&ep_info->dev->dev,"BIDIRECTIONAL MEMCPY CHANNEL %d TEST FAIL %d\n",ch ,ret);
		}
	}
	kfree(buf);
	kfree(validation_buf);
	return ret;
}
#endif
static int set_dma_mask(struct pci_dev *pdev)
{
	int rc;
	if (!pci_set_dma_mask(pdev, DMA_BIT_MASK(64))) {
		rc = pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(64));
		if (rc) {
			rc = pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(32));
			if (rc) {
				dev_err(&pdev->dev,"64-bit DMA enable failed\n");
				return rc;
			}
		}
	} else {
		rc = pci_set_dma_mask(pdev, DMA_BIT_MASK(32));
		if (rc) {
			dev_err(&pdev->dev, "32-bit DMA enable failed\n");
			return rc;
		}
		rc = pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(32));
		if (rc) {
			dev_err(&pdev->dev,
					"32-bit consistent DMA enable failed\n");
			return rc;
		}
	}
	return 0;
}

static int do_pcie_outbound_test_init(struct pci_dev *pdev,
		struct pci_ep_info *ep_info)
{
	int i;

	for(i = 0; i < NUM_PCI_RES; i++) {
		if (!pci_resource_len(pdev, i))
			continue;

		if (pci_resource_flags(pdev, i) & IORESOURCE_IO) {
			dev_err(&ep_info->dev->dev,"**** Not handling PCI I/O resource\n");
			continue;
		}

		ep_info->bars[i].vaddr = ioremap_nocache(pci_resource_start(pdev, i),
				pci_resource_len(pdev, i));
		if (!ep_info->bars[i].vaddr) {
			dev_err(&ep_info->dev->dev,"Unable to map memory region %d\n", i);
			goto cleanup;
		}

		ep_info->bars[i].paddr   = (void *)pci_resource_start(pdev, i);
		ep_info->bars[i].len     = pci_resource_len(pdev, i);
		ep_info->bars[i].enabled = 1;
		dev_info(&ep_info->dev->dev,"res[%d].start : 0x%08llx\n", i, pdev->resource[i].start);
		dev_info(&ep_info->dev->dev,"res[%d].end   : 0x%08llx\n", i, pdev->resource[i].end);
	}

	return 0;
cleanup:
	return -1;
}
static int bidir_ob_test(struct pci_ep_info *ep_info)
{
	u32 val;
	int i, j;

	/* Test only for Bar 0 and Bar 2 */
	for (i = 0; i < 4; i += 2) {
		void *addr = ep_info->bars[i].vaddr;
		if (ep_info->bars[i].enabled) {
			for (j = 0; j < 0x10; j += 4) {
				pr_debug("0x%llx: Write 0x%08x ",
						((long long unsigned int)addr + j), ~j);
				writel(~j, (addr + j));
				val = readl(addr + j);
				if (val != ~j) {
					dev_err(&ep_info->dev->dev,"BAR%d: Outbound Test Fail %X\n", i, val);
					return 1;
				}
			}
		}
		dev_info(&ep_info->dev->dev,"BAR%d: Outbound Test Pass\n", i);
	}
	return 0;
}

static int do_outbound_test(struct pci_ep_info *ep_info)
{
	switch (dma_test)
	{
		case DMA_BIDIRECTIONAL:
			return bidir_ob_test(ep_info);

		case DMA_TO_DEVICE:
			return ob_perf_test(ep_info, ep_info->bars[0].vaddr, 8, DMA_TO_DEVICE, 0);

		case DMA_FROM_DEVICE:
			return ob_perf_test(ep_info, ep_info->bars[0].vaddr, 8, DMA_FROM_DEVICE, 0);

		default:
			return 1;
	}
}

#ifdef CONFIG_XGENE_PCIE_DMA
static struct dma_addr_map *map = NULL;
#endif

static void xgene_pcie_dma_test_start(struct pci_ep_info *ep_info)
{
	switch(test_type) {
		case NORMAL_TEST:
			if(do_outbound_test(ep_info)) {
				pr_err("Outbound Test Error\n");
			} else {
				pr_info("Outbound Test Pass\n");
			}
			break;
#ifdef CONFIG_XGENE_PCIE_DMA
		case MEMCPY_TEST:
			if (do_pcie_dma_memcpy(ep_info)) {
				pr_err("Memcpy Test Error\n");
			} else {
				pr_info("Memcpy Test Pass\n");
			}
			break;
		case GATHER_TEST:
			if (do_pcie_dma_gather(ep_info)) {
				pr_err("Gather DMA Test Error\n");
			} else {
				pr_info("Gather DMA Test Pass\n");
			}
			break;
		case SCATTER_GATHER_TEST:
			if (do_pcie_dma_sg(ep_info)) {
				pr_err("Scatter-Gather DMA Test Error\n");
			} else {
				pr_info("Scatter-Gather DMA Test Pass\n");
			}
			break;
		case EP_DMA_WRITE:
		case EP_DMA_READ:
			{
				int dir = (test_type == EP_DMA_WRITE) ?
					DMA_MEM_TO_DEV : DMA_DEV_TO_MEM;
				struct xgene_pcie_ep_dma *ptr = 
					(struct xgene_pcie_ep_dma *)ep_info->bars[0].vaddr;
				u32 crc = 0;
				map = xgene_pcie_dma_alloc_mem(
						&ep_info->dev->dev,
						0, dir, 0);
				writeq((map[0].phys), &ptr->base);
				writel(map[0].len, &ptr->len);
				writel(dir, &ptr->cmd);
				writel(0, &ptr->status);
				if (test_type == EP_DMA_WRITE)
					crc = crc16(0, map[0].virt, map[0].len);
				writel(crc, &ptr->crc);
			}
			break;
		case EP_DMA_VERIFY:
			{
				struct xgene_pcie_ep_dma *ptr = 
					(struct xgene_pcie_ep_dma *)ep_info->bars[0].vaddr;

				if (ptr->cmd == DMA_DEV_TO_MEM) {
					if (xgene_mem_crc(map) == ptr->crc) {
						pr_info ("\nEP DMA READ Passed\n");	
					}
				} else {
					pr_info ("\nEP DMA WRITE %s\n",
					ptr->status ? "Passed" : "Failed");
				}
			}
			break;
#endif
		default:
			pr_err("Undefined Test Type\n");
	}
}

static int endpoint_probe(struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	struct pci_ep_info *ep_info;
	int rc = 0, i;

	dev_dbg(&pdev->dev,
			"Probing PCIe device [0x%02X:0x%02X]\n",
			pdev->vendor, pdev->device);

	ep_info = kzalloc(sizeof(*ep_info), GFP_KERNEL);
	if(!ep_info)
		return rc;

	ep_info->irq = pdev->irq;
	ep_info->dev = pdev;
	rc = pcim_enable_device(pdev);
	if (rc) { 
		dev_err(&pdev->dev,"unable to enable pci device\n");
		return -ENODEV;
	}
	pci_set_master(pdev);

	dev_dbg (&pdev->dev, "Waiting for Event probe to finish\n");
	xgene_pcievent_init_wait();
	xgene_pcie_gen_msi(1);

#if defined(CONFIG_PCI_MSI)
	rc = pci_enable_msi(pdev);
	if (rc) {
		dev_err(&pdev->dev,"Enable PCIe MSI failed - %d!!\n", rc);
		return -ENODEV;
	}
	dev_dbg(&pdev->dev,"Registering ISR for IRQ %d\n", pdev->irq);
	rc = request_irq(pdev->irq,xgene_event_isr , IRQF_SHARED, PCIE_TEST, ep_info);
	if(rc)
		dev_err(&pdev->dev,"Request IRQ failed!\n");
#endif
	pci_set_drvdata(pdev, ep_info);

	rc = pci_request_regions(pdev, PCIE_TEST);
	if (rc) {
		dev_err(&pdev->dev,"Unable to request memory regions\n");
		return -ENODEV;
	}

	if (do_pcie_outbound_test_init(ep_info->dev, ep_info)) {
		dev_err(&pdev->dev,"Outound init failed\n");
		return -ENODEV;
	}
	/* configure and activate the device */
	rc = set_dma_mask(pdev);

#ifdef CONFIG_XGENE_PCIE_DMA
	xgene_pciedma_hw_init();
	for (i = 0; i < MAX_DMA_CHAN; i++) {
		ep_info->chan[i]=xgene_pcie_dma_request_channel(i);
		if(!ep_info->chan[i]) {
			dev_err(&pdev->dev,"PCIe DMA tests are disabled!\n");
			DMA_DISABLED = 1;
		}
	}
	init_dma_proc(ep_info);
#endif

	/* Do rest init after dma and ep init */
	pr_info ("Endpoint probe complete\n\n");
	return 0;
}

static void endpoint_remove(struct pci_dev *pdev)
{
	int i;
	struct pci_ep_info *ep_info = pci_get_drvdata(pdev);

	dev_info(&ep_info->dev->dev,
			"Removing PCIe device [0x%02X:0x%02X]\n",
			pdev->vendor, pdev->device);
	for (i = 0; i < MAX_DMA_CHAN; i++)
		if (ep_info->chan[i])
			dma_release_channel(ep_info->chan[i]);

	for (i = 0; i < NUM_PCI_RES; i++)
		if (ep_info->bars[i].enabled)
			iounmap(ep_info->bars[i].vaddr);
	pci_release_regions(ep_info->dev);
#if defined(CONFIG_PCI_MSI)
	free_irq(ep_info->dev->irq, ep_info);
	pci_disable_msi(pdev);
#endif
	proc_remove(ep_info->dma_proc);
	kfree(ep_info);
	pr_info ("Endpoint remove complete\n\n");
}

static DEFINE_PCI_DEVICE_TABLE(endpoint_pci_tbl) = {
	{ 0x10e8, 0xcafe, PCI_ANY_ID, PCI_ANY_ID, 0, 0 },
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, endpoint_pci_tbl);

static struct pci_driver test_driver = {
	.name     = PCIE_TEST,
	.id_table = endpoint_pci_tbl,
	.probe    = endpoint_probe,
	.remove   = endpoint_remove,
};

/*
 *  pcie_test_init - Initializes the PCIE_TEST interface
 */

static int __init pcie_test_init(void)
{
	int ret;

	pr_info ("xgene pcie endpoint test driver\n");
	ret = pci_register_driver(&test_driver);
	if (ret)
		pr_err ("Failed to register driver\n");
	/* Set device type to EP for now and update later to RC if probed */
	return 0;
}

/*
 *  pcie_test_exit - Releases the PCIE_TEST interface
 */
static void __exit pcie_test_exit(void)
{
	pr_info ("cleaning up\n");
	pci_unregister_driver(&test_driver);

}

module_init(pcie_test_init);
module_exit(pcie_test_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Tanmay Inamdar <tinamdar@apm.com>");
MODULE_VERSION(PCIE_TEST_VERSION);
MODULE_DESCRIPTION("PCIE_TEST Interface");

