/*
 * AppliedMicro X-Gene MSI Event Driver
 *
 * Copyright (c) 2014, Applied Micro Circuits Corporation
 * Author: Tanmay Inamdar <tinamdar@apm.com>
 *         Mayuresh Chitale <mchitale@apm.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include <linux/pci.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/of.h>
#include <linux/delay.h>

#define	XGENE_PCIE_VENDORID		0x10E8
#define	XGENE_PCIE_EP_DEVICEID		0xCAFE
#define	EP_MODE				0
#define	RC_MODE				1
#define	NUM_PCI_RES			6
#define PATTERN				0xDEADBEEF
#define OFFSET				0x4

extern void xgene_pciedma_hw_init(void);
static struct xgene_msi_event_info
{
	void __iomem *gen_base;
	void __iomem *term_base;
	void __iomem *bar2_virt;
	int mode;
}info;

static void xgene_pciedma_probe (struct work_struct *work)
{
#ifdef CONFIG_XGENE_PCIE_DMA
	xgene_pciedma_hw_init();
#endif
}
DECLARE_WORK(dma_probe, xgene_pciedma_probe);

irqreturn_t xgene_event_isr(int irq, void *data)
{
	u32 val;

	if (info.mode == EP_MODE) {
		pr_info ("Received Interrupt from RC\n");
		val = readl(info.term_base);
		pr_debug ("addr = 0x%llx, data = 0x%x\n", (u64)info.term_base, val);
		/* 
		 * If we received msg from EP driver on RC host
		 * update status so that dma probe may proceed.
		 */
		if (val == 2)
			schedule_work(&dma_probe);
		
	} else {
		pr_info ("Received interrupt from EP\n");
	}
	return IRQ_HANDLED;
}
EXPORT_SYMBOL(xgene_event_isr);

void xgene_pcie_gen_msi (int msg)
{
	u32 val;
	
	if (info.mode == EP_MODE) {
		void __iomem *base = info.gen_base;
		base = info.gen_base;
		writel(1, base); /* enable MSI mode */
		val = readl(base);
		writel(0, base + 4); /* enable MSI sources */
		val = readl(base + 4);
		val = readl(base + 0x100000); /* INT status */
		writel(1, base + 0x200000); /* signal message interrupt pending */
		val = readl(base + 0x200000);
		writel(1, base + 0xa00000); /* clear int status */
		val = readl(base + 0xa00000);
		pr_info("*** Generate MSI from EP to RC\n");
	} else {
		writel(msg, info.bar2_virt);
		pr_info("*** Generate MSI from RC to EP %d\n", msg);
	}
}
EXPORT_SYMBOL(xgene_pcie_gen_msi);
static void xgene_pcievent_probe (void)
{
	struct pci_dev *ep = NULL;

	ep = pci_get_device(XGENE_PCIE_VENDORID, XGENE_PCIE_EP_DEVICEID, NULL);
	if (!ep) {
		pr_err ("Could not find ep device. Event init failed\n");
		return;
	}
	info.mode = RC_MODE;
	if (info.bar2_virt)
		iounmap(info.bar2_virt);
	info.bar2_virt = ioremap_nocache(pci_resource_start(ep, 4),
			pci_resource_len(ep, 4));
	dev_dbg (&ep->dev, "Event Probe done.");
}

void xgene_pcievent_init_wait(void)
{
	xgene_pcievent_probe();
}
EXPORT_SYMBOL(xgene_pcievent_init_wait);

static int xgene_msievent_probe (struct platform_device *pdev)
{
	struct resource *res = NULL;
	struct device_node *np = of_node_get(pdev->dev.of_node);
	const u8 *val;

	val = of_get_property(np, "device_type", NULL);
	if(!val || strcmp(val, "ep"))
		return 0;

	dev_info (&pdev->dev, "EP MSI Event Driver\n");
	info.mode = EP_MODE;
	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "msi_gen");
	info.gen_base = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(info.gen_base)) {
		return PTR_ERR(info.gen_base);
	}

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM, "msi_term");
	info.term_base = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(info.term_base))
		return PTR_ERR(info.term_base);

	res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (!res) {
		dev_err(&pdev->dev, "No irq resource\n");
		return -EINVAL;
	}

	if (devm_request_irq(&pdev->dev, res->start, xgene_event_isr, 0,
				"pcie-misc", &info)) {
		dev_err(&pdev->dev, "request_irq failed.\n");
		return -EINVAL;
	}
	return 0;
}

static const struct of_device_id xgene_msievent_match_table[] = {
	{.compatible = "xgene,xgene-msievent",},
	{},
};

static struct platform_driver xgene_msievent_driver = {
	.driver = {
		   .name = "xgene_msievent",
		   .owner = THIS_MODULE,
		   .of_match_table = of_match_ptr(xgene_msievent_match_table),
	},
	.probe = xgene_msievent_probe,
};
module_platform_driver(xgene_msievent_driver);
